from pm4py.objects.bpmn.exporter.variants.etree import get_xml_string

import powl.conversion.variants.to_bpmn as pminer_bpmn_transformer
import powl.visualization.bpmn.resource_utils.layouter as utils


def apply(activity_to_pool_lane: dict[str, tuple[str, str]], powl) -> str:
    original_pools = set(
        [p for _, (p, _) in activity_to_pool_lane.items() if p is not None]
    )
    activity_to_pool_lane_copy = activity_to_pool_lane.copy()
    if len(original_pools) == 0:
        # Modify it to have a default pool
        activity_to_pool_lane = {
            activity: ("ProcessPool", lane)
            for activity, (_, lane) in activity_to_pool_lane.items()
        }
    for activity, (pool, lane) in activity_to_pool_lane_copy.items():
        if pool is None and lane is None:
            activity_to_pool_lane[activity] = ("ProcessPool", "DefaultLane")
        elif pool is None:
            activity_to_pool_lane[activity] = ("ProcessPool", lane)
        elif lane is None:
            activity_to_pool_lane[activity] = (pool, "DefaultLane")

    pools = utils.__pools_to_tasks(activity_to_pool_lane)
    _, G, _ = pminer_bpmn_transformer.apply(powl)
    coloring = {}
    if len(pools) > 1:
        coloring = utils.color_graph(G, pools)
        G, coloring = utils.__add_intermediate_events_to_graph(G, coloring)
    else:
        pool_name = list(pools.keys())[0] if pools else "DefaultPool"
        coloring = {node: pool_name for node in G.nodes()}
    bpmn, elements_mapping = pminer_bpmn_transformer.__transform_to_bpmn(G)

    # to string
    bpmn = str(get_xml_string(bpmn))
    bpmn = utils.apply_layouting(bpmn)
    bpmn = utils.parse_xml(bpmn)

    # Hardfix coloring keys, don't ask why and how
    # if it includes _, keep it as it is, else add Task_ at the beginning
    coloring = {str(k): v for k, v in coloring.items()}
    coloring = {
        elements_mapping.get(k, k): v
        for k, v in coloring.items()
        if k in elements_mapping
    }
    task_name_to_id = utils.task_name_to_id(bpmn)
    ordered_lanes_and_pools = utils.order_lanes_and_pools(
        activity_to_pool_lane, task_name_to_id, bpmn
    )

    model_dims = utils.get_model_dimensions(bpmn)
    pools = utils.construct_pools(
        activity_to_pool_lane, model_dims, ordered_lanes_and_pools
    )
    lanes = [pool.get_lanes() for pool in pools]
    lanes = [lane for sublist in lanes for lane in sublist]  # Flatten
    task_name_to_id = {k: v for k, v in task_name_to_id.items() if k != ""}
    bpmn, aligned_elements = utils.__align_tasks(lanes, bpmn, task_name_to_id)
    bpmn = utils.__align_elements(bpmn, coloring, aligned_elements, lanes)
    # bpmn = utils.__postprocess_pools(pools, bpmn)
    bpmn = utils.postprocess_diagram(bpmn, pools=pools)
    shapes = utils.__create_shapes(aligned_elements, bpmn)

    # Handle the sequence flows
    bpmn, msg_flows = utils.__handle_sequence_flows(bpmn, shapes)
    bpmn = utils.build_pools_with_collaboration(bpmn, pools, msg_flows)
    return bpmn



from collections import Counter
from multiprocessing import Manager, Pool
from typing import Any, Dict, List, Optional, Tuple

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.strict_tau_loop import (
    StrictTauLoopUVCL,
)
from pm4py.objects.process_tree.obj import Operator

from powl.objects.obj import OperatorPOWL


class POWLStrictTauLoopUVCL(StrictTauLoopUVCL):
    @classmethod
    def apply(
        cls,
        obj: IMDataStructureUVCL,
        pool: Pool = None,
        manager: Manager = None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[OperatorPOWL, List[IMDataStructureUVCL]]]:
        log = obj.data_structure
        proj = cls._get_projected_log(log)
        if sum(proj.values()) > sum(log.values()):
            return OperatorPOWL(Operator.LOOP, []), [
                IMDataStructureUVCL(proj),
                IMDataStructureUVCL(Counter()),
            ]



from abc import ABC
from collections.abc import Collection
from typing import Any, Dict, Generic, List, Optional

from pm4py.algo.discovery.inductive.cuts.loop import LoopCut, LoopCutDFG, LoopCutUVCL, T
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.objects.dfg.obj import DFG
from pm4py.objects.process_tree.obj import Operator

from powl.objects.obj import OperatorPOWL


class POWLLoopCut(LoopCut, ABC, Generic[T]):
    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> OperatorPOWL:
        return OperatorPOWL(Operator.LOOP, [])


class POWLLoopCutUVCL(LoopCutUVCL, POWLLoopCut[IMDataStructureUVCL]):
    pass


class POWLLoopCutDFG(LoopCutDFG, POWLLoopCut[IMDataStructureDFG]):
    @classmethod
    def holds(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[List[Collection[Any]]]:
        """
        This method finds a loop cut in the dfg.
        Implementation follows function LoopCut on page 190 of
        "Robust Process Mining with Guarantees" by Sander J.J. Leemans (ISBN: 978-90-386-4257-4)

        Basic Steps:
        1. merge all start and end activities in one group ('do' group)
        2. remove start/end activities from the dfg
        3. detect connected components in (undirected representative) of the reduced graph
        4. check if each component meets the start/end criteria of the loop cut definition (merge with the 'do' group if not)
        5. return the cut if at least two groups remain

        """
        dfg = obj.dfg
        start_activities = set(dfg.start_activities.keys())
        end_activities = set(dfg.end_activities.keys())
        if len(dfg.graph) == 0:
            return None

        groups = [start_activities.union(end_activities)]
        for c in cls._compute_connected_components(
            dfg, start_activities, end_activities
        ):
            groups.append(set(c.nodes))

        groups = cls._exclude_sets_non_reachable_from_start(
            dfg, start_activities, end_activities, groups
        )
        groups = cls._exclude_sets_no_reachable_from_end(
            dfg, start_activities, end_activities, groups
        )
        groups = cls._check_start_completeness(
            dfg, start_activities, end_activities, groups
        )
        groups = cls._check_end_completeness(
            dfg, start_activities, end_activities, groups
        )

        groups = list(filter(lambda g: len(g) > 0, groups))

        if len(groups) <= 1:
            return None

        return groups

    @classmethod
    def project(
        cls,
        obj: IMDataStructureDFG,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureDFG]:
        dfg = obj.dfg

        skippable = [False for i in range(len(groups))]

        do = groups[0]

        dfgs = [DFG() for i in range(len(groups))]
        activity_to_group_id = {}
        for i in range(len(groups)):
            for a in groups[i]:
                activity_to_group_id[a] = i

        for a in dfg.start_activities:
            if a not in do:
                raise Exception("Start activities must be in the do part of the loop!")
            dfgs[0].start_activities[a] = 1

        for a in dfg.end_activities:
            if a not in do:
                raise Exception("End activities must be in the do part of the loop!")
            dfgs[0].end_activities[a] = 1

        for (a, b) in dfg.graph:

            i = activity_to_group_id[a]
            j = activity_to_group_id[b]

            if i == 0 and j == 0:
                dfgs[0].graph[(a, b)] = dfg.graph[(a, b)]
                # No need for this: if we have a case where the do part is followed by another execution of the do part, then we won't be able to split the execution during projection, and therefore, there this will be modeled as a self-loop in the do part
                # if a in dfg.end_activities and b in dfg.start_activities:
                # skippable[1] = True
            elif i > 0 and j > 0:
                if i == j:
                    dfgs[i].graph[(a, b)] = dfg.graph[(a, b)]
                else:
                    raise Exception("Direct edges between different redo groups!")
            elif (i == 0 and j > 0) or (i > 0 and j == 0):
                dfgs[i].end_activities[a] = 1
                dfgs[j].start_activities[b] = 1
            else:
                raise Exception("We should never reach here!")

        return [
            IMDataStructureDFG(InductiveDFG(dfg=dfgs[i], skip=skippable[i]))
            for i in range(len(dfgs))
        ]



from typing import Any, Dict, List, Optional, Tuple, Type

from pm4py.algo.discovery.inductive.fall_through.empty_traces import EmptyTracesDFG

from powl.discovery.dfg_based.variants.dfg_im_tree import DFGIMBasePOWL, T
from powl.discovery.total_order_based.inductive.fall_through.decision_graph.empty_traces_decision_graph import (
    POWLEmptyTracesDecisionGraphDFG,
)
from powl.discovery.total_order_based.inductive.variants.decision_graph.factory_dg_max import (
    CutFactoryPOWLDecisionGraphMaximal,
)
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.obj import POWL


class DFGPOWLInductiveMinerDecisionGraphMaximal(DFGIMBasePOWL):
    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.DECISION_GRAPH_MAX

    def empty_traces_cut(self) -> Type[EmptyTracesDFG]:
        return POWLEmptyTracesDecisionGraphDFG

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLDecisionGraphMaximal.find_cut(obj, parameters=parameters)
        return res



# this file was initially copied from https://github.com/Nik314/DF2-Miner


def keep_most_frequent_activities(relations, coverage=1.0):

    if coverage >= 1.0:
        return relations

    counts = (
        relations.groupby("ocel:activity")
        .apply(lambda frame: frame["ocel:eid"].nunique())
        .to_dict()
    )
    shares = {key: value / (sum(counts.values())) for key, value in counts.items()}
    shares_sorted = list(reversed(sorted(shares.items(), key=lambda x: x[1])))
    cut_off = min(
        [
            i
            for i in range(len(shares_sorted))
            if sum([entry[1] for entry in shares_sorted[:i]]) >= coverage
        ]
    )
    activities = [shares_sorted[i][0] for i in range(cut_off)]
    return relations[relations["ocel:activity"].isin(activities)]



from typing import List, Tuple


class Lane:
    def __init__(
        self,
        up_left: Tuple[int, int],
        down_right: Tuple[int, int],
        name: str,
        activities: List[str],
    ):
        self.up_left = up_left
        self.down_right = down_right
        self.name = name
        self.activities = activities
        self.elements = []

    def get_up_left(self) -> Tuple[int, int]:
        return self.up_left

    def set_up_left(self, up_left: Tuple[int, int]) -> None:
        self.up_left = up_left

    def get_down_right(self) -> Tuple[int, int]:
        return self.down_right

    def set_down_right(self, down_right: Tuple[int, int]) -> None:
        self.down_right = down_right

    def get_name(self) -> str:
        return self.name

    def get_activities(self) -> List[str]:
        return self.activities

    def get_elements(self) -> List[str]:
        return self.elements

    def add_element(self, element: str) -> None:
        self.elements.append(element)

    def has_element(self, element: str) -> bool:
        return element in self.elements



from typing import Any, Dict, Optional

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)

from powl.discovery.total_order_based.inductive.base_case.abc import BaseCase

from powl.objects.obj import Transition


class SingleActivityBaseCaseUVCL(BaseCase[IMDataStructureUVCL]):
    @classmethod
    def holds(
        cls, obj=IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> bool:
        if len(obj.data_structure.keys()) != 1:
            return False
        if len(list(obj.data_structure.keys())[0]) > 1:
            return False
        return True

    @classmethod
    def leaf(
        cls, obj=IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> Transition:
        for t in obj.data_structure:
            if t:
                return Transition(label=t[0])
            else:
                return Transition()


class SingleActivityBaseCaseDFG(BaseCase[IMDataStructureDFG]):
    @classmethod
    def holds(
        cls,
        obj=IMDataStructureDFG,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> bool:
        if len(obj.dfg.graph) == 0:
            if set(obj.dfg.start_activities) == set(obj.dfg.end_activities):
                return len(obj.dfg.start_activities) == 1
            else:
                raise Exception(
                    "Invalid DFG: non-start/end activities are not involved in any edges!"
                )
        return False

    @classmethod
    def leaf(
        cls,
        obj=IMDataStructureDFG,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Transition:
        return Transition(label=list(obj.dfg.start_activities)[0])



import copy
import importlib.resources
import os
import tempfile
from enum import Enum

import graphviz
from graphviz import Digraph
from pm4py.objects.bpmn.obj import BPMN
from pm4py.objects.bpmn.util import reduction
from pm4py.objects.bpmn.util.sorting import get_sorted_nodes_edges
from pm4py.util import constants

from powl.conversion.variants.to_petri_net import apply as powl_to_pn

from powl.objects.obj import POWL


class Parameters(Enum):
    FORMAT = "format"
    RANKDIR = "rankdir"
    BGCOLOR = "bgcolor"


FREQUENCY_TAG_IMAGES = True
SPLIT_EXCLUSIVE_GATEWAYS = True
min_width = "1.5"  # Set the minimum width in inches
min_height = "0.5"
node_color = "#f2f2f2"
silent_node_color = "black"


class SplitExclusiveGateway(BPMN.ExclusiveGateway):
    pass


class JoinExclusiveGateway(BPMN.ExclusiveGateway):
    pass


def can_connect_without_xor_violation(input_nodes, output_nodes):
    for in_node in input_nodes:
        for out_node in output_nodes:
            if isinstance(in_node, BPMN.ExclusiveGateway) and len(output_nodes) > 1:
                return False
            if isinstance(out_node, BPMN.ExclusiveGateway) and len(input_nodes) > 1:
                return False

    return True


def remove_all_occurrences_from_list(item, items):
    return [other for other in items if other is not item]


def simplify_and_gateways(nodes, edges):
    for n in list(nodes):
        if isinstance(n, BPMN.ParallelGateway):
            input_nodes = [edge[0] for edge in edges if edge[1] == n]
            output_nodes = [edge[1] for edge in edges if edge[0] == n]
            if can_connect_without_xor_violation(input_nodes, output_nodes):
                nodes = remove_all_occurrences_from_list(n, nodes)
                edges = [edge for edge in edges if n not in edge]

                for in_node in input_nodes:
                    for out_node in output_nodes:
                        edges.append((in_node, out_node))
                return simplify_and_gateways(nodes, edges)

    return nodes, edges


def add_node(n, viz):
    n_id = str(id(n))
    if isinstance(n, FrequencyTask):
        font_size = "18"
        if FREQUENCY_TAG_IMAGES:
            if n.skippable:
                if n.selfloop:
                    with importlib.resources.path(
                        "powl.visualization.powl.variants.icons", "skip-loop-tag.svg"
                    ) as gimg:
                        image = str(gimg)
                        viz.node(
                            n_id,
                            label="\n" + n.activity,
                            imagepos="tr",
                            image=image,
                            style="filled",
                            fillcolor=node_color,
                            shape="box",
                            width=min_width,
                            fontsize=font_size,
                        )
                else:
                    with importlib.resources.path(
                        "powl.visualization.powl.variants.icons", "skip-tag.svg"
                    ) as gimg:
                        skip_image = str(gimg)
                        viz.node(
                            n_id,
                            label="\n" + n.activity,
                            imagepos="tr",
                            image=skip_image,
                            style="filled",
                            fillcolor=node_color,
                            shape="box",
                            width=min_width,
                            fontsize=font_size,
                        )
            else:
                if n.selfloop:
                    with importlib.resources.path(
                        "powl.visualization.powl.variants.icons", "loop-tag.svg"
                    ) as gimg:
                        loop_image = str(gimg)
                        viz.node(
                            n_id,
                            label="\n" + n.activity,
                            imagepos="tr",
                            image=loop_image,
                            style="filled",
                            fillcolor=node_color,
                            shape="box",
                            width=min_width,
                            fontsize=font_size,
                        )
                else:
                    viz.node(
                        n_id,
                        label=n.activity,
                        style="filled",
                        fillcolor=node_color,
                        shape="box",
                        width=min_width,
                        fontsize=font_size,
                    )
        else:
            if n.skippable:
                viz.node(
                    n_id,
                    shape="box",
                    label=n.activity,
                    fontsize=font_size,
                    style="dashed",
                    width=min_width,
                )
            else:
                viz.node(
                    n_id,
                    shape="box",
                    label=n.activity,
                    fontsize=font_size,
                    width=min_width,
                )
            if n.selfloop:
                viz.edge(n_id, n_id)

    elif isinstance(n, BPMN.StartEvent):
        with importlib.resources.path(
            "powl.visualization.powl.variants.icons", "play.svg"
        ) as gimg:
            start_image = str(gimg)
            viz.node(
                n_id,
                image=start_image,
                label="",
                shape="none",
                width="0.35",
                height="0.35",
                fixedsize="true",
                style="filled",
                fillcolor=node_color,
            )
    elif isinstance(n, BPMN.EndEvent):
        with importlib.resources.path(
            "powl.visualization.powl.variants.icons", "end.svg"
        ) as gimg:
            end_image = str(gimg)
            viz.node(
                n_id,
                image=end_image,
                label="",
                shape="none",
                width="0.35",
                height="0.35",
                fixedsize="true",
                style="filled",
                fillcolor=node_color,
            )
    elif isinstance(n, BPMN.ParallelGateway):
        viz.node(
            n_id,
            label="",
            shape="square",
            style="filled",
            fillcolor=silent_node_color,
            width="0.3",
            height="0.3",
        )
    elif isinstance(n, SplitExclusiveGateway):
        with importlib.resources.path(
            "powl.visualization.powl.variants.icons", "gate.svg"
        ) as gimg:
            xor_image = str(gimg)
            viz.node(
                n_id,
                label="",
                shape="diamond",
                style="filled",
                fillcolor="lightgreen",
                width="0.4",
                height="0.4",
                fixedsize="true",
                image=xor_image,
            )
    elif isinstance(n, JoinExclusiveGateway):
        with importlib.resources.path(
            "powl.visualization.powl.variants.icons", "gate.svg"
        ) as gimg:
            xor_image = str(gimg)
            viz.node(
                n_id,
                label="",
                shape="diamond",
                style="filled",
                fillcolor="orange",
                width="0.4",
                height="0.4",
                fixedsize="true",
                image=xor_image,
            )
    elif isinstance(n, BPMN.ExclusiveGateway):
        with importlib.resources.path(
            "powl.visualization.powl.variants.icons", "gate.svg"
        ) as gimg:
            xor_image = str(gimg)
        viz.node(
            n_id,
            label="",
            shape="diamond",
            style="filled",
            fillcolor=node_color,
            width="0.4",
            height="0.4",
            fixedsize="true",
            image=xor_image,
        )
    else:
        raise Exception("Unexpected instance of class " + str(type(n)) + "!")


def split_xor_gateways(nodes, edges):
    for node in list(nodes):
        if isinstance(node, BPMN.ExclusiveGateway):
            preset = [e[0] for e in edges if e[1] is node]
            postset = [e[1] for e in edges if e[0] is node]

            if len(preset) == 1 and len(postset) == 1:
                edges = remove_all_occurrences_from_list((preset[0], node), edges)
                edges = remove_all_occurrences_from_list((node, postset[0]), edges)
                edges.append((preset[0], postset[0]))

            elif len(preset) == 1 and len(postset) > 1:
                node.__class__ = SplitExclusiveGateway

            elif len(preset) > 1 and len(postset) == 1:
                node.__class__ = JoinExclusiveGateway

            elif len(preset) > 1 and len(postset) > 1:
                join_node = copy.deepcopy(node)
                join_node.__class__ = JoinExclusiveGateway
                nodes.append(join_node)

                split_node = copy.deepcopy(node)
                split_node.__class__ = SplitExclusiveGateway
                nodes.append(split_node)
                nodes = remove_all_occurrences_from_list(node, nodes)

                for idx, edge in enumerate(edges):
                    if edge[1] is node:
                        edges[idx] = (edge[0], join_node)
                    if edge[0] is node:
                        edges[idx] = (split_node, edge[1])

                edges.append((join_node, split_node))
    return nodes, edges


def apply(powl: POWL) -> graphviz.Digraph:
    pn_2, init_2, final_2 = powl_to_pn(
        powl, parameters={"flatten_frequency_tags": False}
    )
    bpmn_graph = to_bpmn(pn_2, init_2, final_2)

    nodes, edges = get_sorted_nodes_edges(bpmn_graph)

    nodes, edges = simplify_and_gateways(nodes, edges)

    rankdir = "LR"
    bgcolor = constants.DEFAULT_BGCOLOR

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    viz = Digraph(
        "", filename=filename.name, engine="dot", graph_attr={"bgcolor": bgcolor}
    )
    viz.graph_attr["rankdir"] = rankdir

    gateway_edges = {}

    for e in edges:
        if isinstance(e[0], BPMN.ParallelGateway) or isinstance(
            e[0], BPMN.ParallelGateway
        ):
            if e[0] not in gateway_edges:
                gateway_edges[e[0]] = {"in": [], "out": []}
            gateway_edges[e[0]]["out"].append(e[1])
            continue
        if isinstance(e[1], BPMN.ParallelGateway) or isinstance(
            e[1], BPMN.ParallelGateway
        ):
            if e[1] not in gateway_edges:
                gateway_edges[e[1]] = {"in": [], "out": []}
            gateway_edges[e[1]]["in"].append(e[0])
            continue

    if SPLIT_EXCLUSIVE_GATEWAYS:
        nodes, edges = split_xor_gateways(nodes, edges)

    for node in nodes:
        add_node(node, viz)
    add_concurrent_subgraphs(viz, find_concurrent_groups(nodes, edges))

    for e in edges:
        n_id_1 = str(id(e[0]))
        n_id_2 = str(id(e[1]))

        viz.edge(n_id_1, n_id_2)

    viz.attr(overlap="false")

    viz.format = "svg"

    return viz


class FrequencyTask(BPMN.Task):
    def __init__(
        self, name, properties, id="", in_arcs=None, out_arcs=None, process=None
    ):
        super().__init__(
            id=id, name=name, in_arcs=in_arcs, out_arcs=out_arcs, process=process
        )
        self.activity = properties["activity"]
        self.skippable = properties["skippable"]
        self.selfloop = properties["selfloop"]
        script_dir = os.path.dirname(os.path.realpath(__file__))
        self.image = os.path.join(script_dir, "xor2.png")


def to_bpmn(net, im, fm):
    """
    Converts an accepting Petri net into a BPMN diagram

    Parameters
    --------------
    net
        Petri net
    im
        initial marking
    fm
        final marking

    Returns
    --------------
    bpmn_graph
        BPMN diagram
    """

    bpmn_graph = BPMN()

    entering_dictio = {}
    exiting_dictio = {}

    for place in net.places:
        node = BPMN.ExclusiveGateway()
        bpmn_graph.add_node(node)
        entering_dictio[place] = node
        exiting_dictio[place] = node

    for trans in net.transitions:
        if trans.label is None:
            if len(trans.in_arcs) > 1:
                node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.CONVERGING
                )
            elif len(trans.out_arcs) > 1:
                node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.DIVERGING
                )
            else:
                node = BPMN.ExclusiveGateway(
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED
                )
            bpmn_graph.add_node(node)
            entering_dictio[trans] = node
            exiting_dictio[trans] = node
        else:
            if len(trans.in_arcs) > 1:
                entering_node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.CONVERGING
                )
            else:
                entering_node = BPMN.ExclusiveGateway(
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED
                )

            if len(trans.out_arcs) > 1:
                exiting_node = BPMN.ParallelGateway(
                    gateway_direction=BPMN.Gateway.Direction.DIVERGING
                )
            else:
                exiting_node = BPMN.ExclusiveGateway(
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED
                )

            task = FrequencyTask(name=trans.label, properties=trans.properties)
            bpmn_graph.add_node(task)

            bpmn_graph.add_flow(BPMN.Flow(entering_node, task))
            bpmn_graph.add_flow(BPMN.Flow(task, exiting_node))

            entering_dictio[trans] = entering_node
            exiting_dictio[trans] = exiting_node

    for arc in net.arcs:
        bpmn_graph.add_flow(
            BPMN.Flow(exiting_dictio[arc.source], entering_dictio[arc.target])
        )

    start_node = BPMN.StartEvent(name="start", isInterrupting=True)
    end_node = BPMN.NormalEndEvent(name="end")
    bpmn_graph.add_node(start_node)
    bpmn_graph.add_node(end_node)
    for place in im:
        bpmn_graph.add_flow(BPMN.Flow(start_node, entering_dictio[place]))
    for place in fm:
        bpmn_graph.add_flow(BPMN.Flow(exiting_dictio[place], end_node))

    bpmn_graph = reduction.apply(bpmn_graph)

    for node in bpmn_graph.get_nodes():
        node.set_process(bpmn_graph.get_process_id())

    for edge in bpmn_graph.get_flows():
        edge.set_process(bpmn_graph.get_process_id())

    return bpmn_graph


def find_concurrent_groups(nodes, edges):
    predecessors = {n: [] for n in nodes}
    successors = {n: [] for n in nodes}
    for src, dst in edges:
        successors[src].append(dst)
        predecessors[dst].append(src)

    groups = {}
    for node, preds in predecessors.items():
        if len(preds) == 1:
            pred_id = str(id(preds[0]))
            if pred_id not in groups:
                groups[pred_id] = []
            groups[pred_id].append(node)

    return [group for group in groups.values() if len(group) > 1]


def add_concurrent_subgraphs(graph, concurrent_elements):
    for group in concurrent_elements:
        with graph.subgraph() as s:
            s.attr(rank="same")
            for n in group:
                s.node(str(id(n)))



from collections import Counter
from multiprocessing import Manager, Pool
from typing import Any, Dict, List, Optional, Tuple

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL

from pm4py.algo.discovery.inductive.fall_through.activity_concurrent import (
    ActivityConcurrentUVCL,
)

from powl.objects.obj import StrictPartialOrder


class POWLActivityConcurrentUVCL(ActivityConcurrentUVCL):
    @classmethod
    def apply(
        cls,
        obj: IMDataStructureUVCL,
        pool: Pool = None,
        manager: Manager = None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[StrictPartialOrder, List[IMDataStructureUVCL]]]:
        candidate = cls._get_candidate(obj, pool, manager, parameters)
        if candidate is None:
            return None
        log = obj.data_structure
        l_a = Counter()
        l_other = Counter()
        for t in log:
            l_a.update({tuple(filter(lambda e: e == candidate, t)): log[t]})
            l_other.update({tuple(filter(lambda e: e != candidate, t)): log[t]})
        return StrictPartialOrder([]), [
            IMDataStructureUVCL(l_a),
            IMDataStructureUVCL(l_other),
        ]



from typing import Any, Dict, List, Optional, Tuple

from powl.discovery.dfg_based.variants.dfg_im_tree import DFGIMBasePOWL, T
from powl.discovery.total_order_based.inductive.variants.maximal.factory import (
    CutFactoryPOWLMaximal,
)
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.obj import POWL


class DFGPOWLInductiveMinerMaximalOrder(DFGIMBasePOWL):
    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.MAXIMAL

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLMaximal.find_cut(obj, parameters=parameters)
        return res



from typing import Any, Dict, List, Optional, Tuple

from powl.discovery.total_order_based.inductive.variants.decision_graph.factory_cyclic_dg import (
    CutFactoryCyclicDecisionGraph,
)
from powl.discovery.total_order_based.inductive.variants.decision_graph.factory_cyclic_dg_strict import (
    CutFactoryCyclicDecisionGraphStrict,
)
from powl.discovery.total_order_based.inductive.variants.im_decision_graph_maximal import (
    POWLInductiveMinerDecisionGraphMaximal,
)
from powl.discovery.total_order_based.inductive.variants.im_tree import T
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.obj import POWL


class POWLInductiveMinerDecisionGraphCyclic(POWLInductiveMinerDecisionGraphMaximal):
    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.DECISION_GRAPH_CYCLIC

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryCyclicDecisionGraph.find_cut(obj, parameters=parameters)
        return res


class POWLInductiveMinerDecisionGraphCyclicStrict(
    POWLInductiveMinerDecisionGraphMaximal
):
    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.DECISION_GRAPH_CYCLIC_STRICT

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryCyclicDecisionGraphStrict.find_cut(obj, parameters=parameters)
        return res



from abc import ABC
from typing import Any, Dict, Generic, Optional

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)

from powl.discovery.total_order_based.inductive.base_case.abc import BaseCase, T
from powl.objects.obj import SilentTransition


class EmptyLogBaseCase(BaseCase[T], ABC, Generic[T]):
    @classmethod
    def leaf(
        cls, obj=T, parameters: Optional[Dict[str, Any]] = None
    ) -> SilentTransition:
        return SilentTransition()


class EmptyLogBaseCaseUVCL(EmptyLogBaseCase[IMDataStructureUVCL]):
    @classmethod
    def holds(
        cls, obj=IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> bool:
        return len(obj.data_structure) == 0


class EmptyLogBaseCaseDFG(EmptyLogBaseCase[IMDataStructureDFG]):
    @classmethod
    def holds(
        cls,
        obj=IMDataStructureDFG,
        parameters: Optional[Dict[str, Any]] = None,
    ):
        dfg = obj.dfg
        return (
            len(dfg.graph) == 0
            and len(dfg.start_activities) == 0
            and len(dfg.end_activities) == 0
        )



from typing import Any, Dict, Optional, Type, TypeVar

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureDFG

from powl.discovery.total_order_based.inductive.fall_through.empty_traces import (
    POWLEmptyTracesDFG,
)
from powl.discovery.total_order_based.inductive.variants.im_tree import IMBasePOWL
from powl.objects.obj import POWL

T = TypeVar("T", bound=IMDataStructureDFG)


class DFGIMBasePOWL(IMBasePOWL[T]):
    def empty_traces_cut(self) -> Type[POWLEmptyTracesDFG]:
        return POWLEmptyTracesDFG

    def apply(
        self,
        obj: T,
        parameters: Optional[Dict[str, Any]] = None,
        second_iteration: bool = False,
    ) -> POWL:

        empty_traces = self.empty_traces_cut().apply(obj, parameters)
        if empty_traces is not None:
            return self._recurse(empty_traces[0], empty_traces[1], parameters)

        powl = self.apply_base_cases(obj, parameters)
        if powl is not None:
            return powl

        cut = self.find_cut(obj, parameters)
        if cut is not None:
            powl = self._recurse(cut[0], cut[1], parameters=parameters)

        if powl is not None:
            return powl

        ft = self.fall_through(obj, parameters)
        return self._recurse(ft[0], ft[1], parameters=parameters)



from typing import TypeVar

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.statistics.eventually_follows.uvcl.get import apply as to_efg
from pm4py.objects.dfg import util as dfu


T = TypeVar("T", bound=IMDataStructureUVCL)


def filter_efg_based_on_filtered_dfg(obj, alphabet, dfg, filtering_threshold):
    efg = to_efg(obj)
    if filtering_threshold is None:
        return efg
    filtered_efg = {}
    transitive_predecessors, transitive_successors = dfu.get_transitive_relations(
        dfg
    )
    for a in alphabet:
        for b in alphabet:
            if (a, b) in efg:
                if b in transitive_successors[a]:
                    filtered_efg[(a, b)] = efg[(a, b)]
    return filtered_efg


import ast
import itertools
import random
import shutil

import subprocess
from collections import defaultdict

from enum import Enum
from importlib.resources import files
from itertools import product
from typing import Dict, List, Set, Tuple

import networkx as nx
import shapely
from lxml import etree

from powl.visualization.bpmn.resource_utils.lanes import Lane

from powl.visualization.bpmn.resource_utils.pools import Pool

PADDING_LANES = 50
PADDING_POOLS = 60


class DockingDirection(Enum):
    """
    This class is used to define the docking directions for the elements
    """

    LEFT = "l"
    RIGHT = "r"
    TOP = "t"
    BOTTOM = "b"


def __get_gateway_type(G, node):
    predecessors = list(G.predecessors(node))
    successors = list(G.successors(node))
    if len(predecessors) > len(successors):
        return "converging"
    elif len(predecessors) < len(successors):
        return "diverging"
    else:
        return "mixed"


def apply_layouting(content_diagram: str) -> str:
    """
    Apply BPMN auto-layout using the bundled JS script.
    Requires Node.js to be installed and accessible on PATH.
    """

    # Ensure Node.js is available
    node = shutil.which("node")
    if not node:
        raise RuntimeError(
            "Node.js is required but was not found on PATH. "
            "Please install Node.js from https://nodejs.org/."
        )

    # Locate the bundled JS file inside powl/js/
    js_bundle = files("powl") / "js" / "layout.bundle.cjs"

    if not js_bundle.exists():
        raise FileNotFoundError(
            f"Could not find {js_bundle}. "
            "Make sure you ran `npm run build` and copied the bundle into powl/js/."
        )

    # 3. Run Node process with stdin/stdout
    process = subprocess.Popen(
        [node, str(js_bundle)],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    stdout, stderr = process.communicate(input=content_diagram)

    if process.returncode != 0:
        raise RuntimeError(
            f"Auto-layout failed (exit {process.returncode}): {stderr.strip()}"
        )

    return stdout


def __pools_to_tasks(
    activity_to_pool_lane: Dict[str, Tuple[str, str]]
) -> Dict[str, set]:
    """
    Convert pools to tasks based on the lane data.
    """
    pool_tasks = defaultdict(set)
    for activity, (pool, _) in activity_to_pool_lane.items():
        if pool is None:
            if "DefaultPool" not in pool_tasks:
                pool_tasks["DefaultPool"] = set()
            pool_tasks["DefaultPool"].add(activity)
        else:
            if pool not in pool_tasks:
                pool_tasks[pool] = set()
            pool_tasks[pool].add(activity)
    return dict(pool_tasks)


def __color_subgraph(
    graph: nx.DiGraph,
    start_node: str,
    end_node: str,
    coloring: Dict[str, str],
    main_color: str,
) -> nx.DiGraph:
    """
    Generate a subgraph from start_node to end_node.

    :param graph: The directed graph to extract the subgraph from.
    :param start_node: The starting node of the subgraph.
    :param end_node: The ending node of the subgraph.
    :return: A subgraph containing nodes and edges between start_node and end_node.
    """
    list_of_paths = list(nx.all_simple_paths(graph, source=start_node, target=end_node))
    nodes_in_subgraph = set([node for path in list_of_paths for node in path])
    subgraph = graph.subgraph(nodes_in_subgraph).copy()
    # Check for all tasks included there
    used_colors = [
        coloring[node]
        for node in subgraph.nodes
        if node in coloring and graph.nodes[node].get("content") is not None
    ]
    """    print('--- Coloring Subgraph ---')
        print(f"The subgraph nodes between {start_node} and {end_node} are: {nodes_in_subgraph}")
        print(f"The used colors are: {set(used_colors)}")
    """
    if len(set(used_colors)) == 1:
        # Color all nodes in the subgraph with the same color
        color = used_colors[0]
        for node in subgraph.nodes:
            coloring[node] = color

    return coloring


def __get_predecessors_color(G, node, coloring):
    predecessors = list(G.predecessors(node))
    for predecessor in predecessors:
        if predecessor in coloring:
            return coloring[predecessor]
    return None


def __get_successors_color(G, node, coloring):
    successors = list(G.successors(node))
    for successor in successors:
        if successor in coloring:
            return coloring[successor]
    return None


def color_graph(graph: nx.DiGraph, pools: Dict[str, List]) -> List[Tuple[str, str]]:
    """
    Color the graph based on the pools.

    :param graph: The directed graph to color.
    :param pools: A dictionary where keys are pool names and values are sets of node contents.
    :return: A list of tuples with node id and color.
    """
    pool_with_most_tasks = max(pools, key=lambda k: len(pools[k]))
    coloring = {}
    for pool, tasks in pools.items():
        # identify the node with this content
        for task in tasks:
            node_for_task = next(
                (
                    n
                    for n, attr in graph.nodes(data=True)
                    if attr.get("content") == task
                ),
                None,
            )
            if node_for_task:
                # color the node with the pool name
                coloring[node_for_task] = pool
    for node in graph.nodes:
        counterparts = graph.nodes[node].get("paired_with")
        counterparts = (
            [counterpart for counterpart in counterparts if counterpart in graph.nodes]
            if counterparts
            else []
        )
        if graph.nodes[node].get("type") == "diverging":
            counterparts = graph.nodes[node].get("paired_with")
            for counterpart in counterparts:
                if counterpart in graph.nodes:
                    coloring = __color_subgraph(
                        graph, node, counterpart, coloring, pool_with_most_tasks
                    )
    for node in graph.nodes:
        if node not in coloring.keys():
            if "Gateway" in node:
                # get its type
                gateway_type = __get_gateway_type(graph, node)
                if gateway_type == "converging":
                    color = __get_predecessors_color(graph, node, coloring)
                    if color:
                        coloring[node] = color
                elif gateway_type == "diverging":
                    color = __get_successors_color(graph, node, coloring)
                    if color:
                        coloring[node] = color
                if node not in coloring.keys():
                    coloring[node] = pool_with_most_tasks
    for node in graph.nodes:
        if node not in coloring.keys():
            if graph.nodes[node].get("type") == "startEvent":
                # Get the color of the successor
                successors = list(graph.successors(node))
                while successors:
                    current_element = successors.pop(0)
                    if current_element in coloring.keys():
                        coloring[node] = coloring[current_element]
                        break
                    # otherwise, add its successors to the list
                    for succ in graph.successors(current_element):
                        successors.append(succ)
            elif graph.nodes[node].get("type") == "endEvent":
                # Get the color of the predecessor
                predecessors = list(graph.predecessors(node))
                while predecessors:
                    current_element = predecessors.pop(0)
                    if current_element in coloring.keys():
                        coloring[node] = coloring[current_element]
                        break
                    # otherwise, add its predecessors to the list
                    for pred in graph.predecessors(current_element):
                        predecessors.append(pred)

    return dict(coloring)


def __discover_edges_to_drop(
    G: nx.DiGraph, connected_component: Set[str]
) -> List[Tuple[str, str]]:
    """
    Discover edges to drop between components in the graph.

    :param G: The directed graph to discover edges from.
    :param connected_component: A set of node ids representing the submodel.
    :return: A list of edges to drop.
    """
    edges_to_drop = []
    for u, v in G.edges():
        if (u in connected_component and v not in connected_component) or (
            u not in connected_component and v in connected_component
        ):
            edges_to_drop.append((u, v))
    return edges_to_drop


def __add_intermediate_events_to_graph(G: nx.DiGraph, coloring: dict[str, str]):
    G_new = G.copy()
    coloring_new = coloring.copy()
    edges_in_graph = list(G.edges())
    edges_to_add = []
    edges_to_remove = []
    for u, v in edges_in_graph:
        # Check if the colors of the source (u) and destination (v) nodes are different.
        if coloring.get(u) != coloring.get(v):

            # Define unique names for the new intermediate nodes.
            # Using the original node names makes the new names more readable.
            throw_node_id = f"IntermediateThrowEvent_{hash(u)}_{hash(v)}"
            catch_node_id = f"IntermediateCatchEvent_{hash(u)}_{hash(v)}"

            # Add the new "throw" node to the graph and the coloring dictionary.
            # It inherits the color of the source node (u).
            G_new.add_node(throw_node_id, type="throw")
            coloring_new[throw_node_id] = coloring[u]

            # Add the new "catch" node to the graph and the coloring dictionary.
            # It inherits the color of the destination node (v).
            G_new.add_node(catch_node_id, type="catch")
            coloring_new[catch_node_id] = coloring[v]

            # Schedule the original edge for removal.
            edges_to_remove.append((u, v))

            # Schedule the new sequence of edges to be added.
            edges_to_add.append((u, throw_node_id))
            edges_to_add.append((throw_node_id, catch_node_id))
            edges_to_add.append((catch_node_id, v))

    # Apply all the collected changes to the graph at once.
    G_new.remove_edges_from(edges_to_remove)
    G_new.add_edges_from(edges_to_add)
    return G_new, coloring_new


def parse_xml(xml: bytes | str) -> etree._Element:
    root = None
    if isinstance(xml, str) and xml.startswith(("b'", 'b"')):
        try:
            # ast.literal_eval is the safe way to do this. NEVER use eval().
            xml_bytes = ast.literal_eval(xml)
        except (ValueError, SyntaxError):
            raise ValueError(
                f"Failed to parse string-formatted bytes: {xml_bytes[:100]}..."
            )

        parser = etree.XMLParser(remove_blank_text=True, recover=True, encoding="utf-8")
        root = etree.fromstring(xml_bytes, parser=parser)
    else:
        parser = etree.XMLParser(remove_blank_text=True, recover=True)
        root = etree.fromstring(xml.encode("utf-8"), parser)

    # check if root has nsmap
    if not root.nsmap:
        raise ValueError("The provided XML does not contain any namespace mappings.")
    return root


def _to_string(root: etree._Element) -> str:
    return etree.tostring(
        root, pretty_print=True, xml_declaration=True, encoding="UTF-8"
    ).decode("utf-8")


def remove_edge_information(root: etree._Element) -> etree._Element:
    """
    Identifies and removes all BPMNEdge elements from the XML tree
    using dynamically discovered namespaces.
    """
    bpmndi_ns = root.nsmap.get("bpmndi")
    if not bpmndi_ns:
        # Handle case where the namespace is not defined
        raise Exception(
            "BPMNDI namespace not found in the document. No edges will be removed."
        )

    edge_tag = f"{{{bpmndi_ns}}}BPMNEdge"
    edges_to_remove = root.findall(f".//{edge_tag}")

    for edge in edges_to_remove:
        parent = edge.getparent()
        if parent is not None:
            parent.remove(edge)

    return root


def identify_edges_id(root: etree._Element) -> List[str]:
    """
    Identifies the edges in the XML output and returns their 'bpmnElement' IDs
    using the findall method.
    """
    bpmndi_ns = root.nsmap.get("bpmndi")
    if not bpmndi_ns:
        raise Exception(
            "BPMNDI namespace not found in the document. Cannot identify edges."
        )

    edge_tag = f"{{{bpmndi_ns}}}BPMNEdge"
    edges = root.findall(f".//{edge_tag}")

    edge_ids = [edge.get("bpmnElement") for edge in edges]

    return [eid for eid in edge_ids if eid is not None]


def get_model_dimensions(root: etree._Element, padding: float = 30) -> List[int]:
    """
    Get the dimensions of the model.
    """
    # dc:Bounds x="" and whatever afterwards
    OMGDC_NS = root.nsmap.get("omgdc")

    # For compatibility reasons, also check for "dc" namespace
    if not OMGDC_NS:
        OMGDC_NS = root.nsmap.get("dc")

    if not OMGDC_NS:
        raise Exception(
            "(OMG)DC namespace not found in the document. Cannot determine model dimensions."
        )

    boundaries_tag = f"{{{OMGDC_NS}}}Bounds"
    all_bounds = root.findall(f".//{boundaries_tag}")

    if not all_bounds:
        raise Exception(
            "No Bounds found in the document. Cannot determine model dimensions."
        )

    max_right_edge = 0.0
    max_bottom_edge = 0.0
    max_height = 0.0

    for bound in all_bounds:
        if not all(attr in bound.attrib for attr in ["x", "y", "width", "height"]):
            raise Exception("Bounds element missing required attributes.")

        x = float(bound.get("x"))
        y = float(bound.get("y"))
        width = float(bound.get("width"))
        height = float(bound.get("height"))
        if height > max_height:
            max_height = height
        max_right_edge = max(max_right_edge, x + width)
        max_bottom_edge = max(max_bottom_edge, y + height)

    return [
        int(max_right_edge + padding),
        int(max_bottom_edge + padding),
        int(max_height),
    ]


def task_name_to_id(root: etree._Element) -> dict:
    """
    This function takes the xml output and the name of the element and returns the id of the element
    """
    BPMN_NS = root.nsmap.get("bpmn")
    if not BPMN_NS:
        raise Exception(
            "BPMN namespace not found in the document. Cannot map task names to IDs."
        )

    name_to_id = {}

    bpmn_tag = f"{{{BPMN_NS}}}"

    # Tags we want to find
    tags_to_find = [
        "task",
        "startEvent",
        "endEvent",
        "intermediateThrowEvent",
        "intermediateCatchEvent",
    ]
    results_iterators = [root.iterfind(f".//{bpmn_tag}{tag}") for tag in tags_to_find]

    all_elements = itertools.chain.from_iterable(results_iterators)

    for element in all_elements:
        element_id = element.get("id")
        if not element_id:
            continue
        key_name = element.get("name") or element_id
        name_to_id[key_name] = element_id

    return name_to_id


def __get_element_coordinates(
    root: etree._Element, element_id: str
) -> Tuple[float, float]:
    """
    This function takes the xml output and the id of the activity and returns the coordinates of the activity
    """
    ns = root.nsmap
    if "bpmndi" not in ns or "dc" not in ns:
        # Some tools use 'omgdc' as an alias for the 'dc' namespace URI
        if "omgdc" in ns and "dc" not in ns:
            ns["dc"] = ns["omgdc"]
        else:
            raise Exception("BPMNDI or DC namespaces not found.")

    query = f'.//bpmndi:BPMNShape[@bpmnElement="{element_id}"]/dc:Bounds'
    bounds_element = root.find(query, namespaces=ns)
    if bounds_element is not None:
        x = float(bounds_element.get("x"))
        y = float(bounds_element.get("y"))
        width = float(bounds_element.get("width"))
        height = float(bounds_element.get("height"))
        return (x, y, width, height)
    else:
        raise Exception(f"Element with id {element_id} not found in the XML output.")


def __order_structures(
    data: dict, task_name_to_id: dict, root: etree._Element
) -> List[str]:
    sorting = {}
    for structure in data.keys():
        min_x = float("inf")
        activities = data[structure]
        for activity in activities:
            activity_id = task_name_to_id.get(activity)
            if activity_id:
                x = __get_element_coordinates(root, activity_id)[0]
                if x < min_x:
                    min_x = x
        sorting[structure] = min_x
    # Sort them based on the minimal x
    ordered_structures = sorted(sorting, key=sorting.get)
    return ordered_structures


def order_lanes_and_pools(
    activity_to_pool_lane: Dict[str, Tuple[str, str]],
    task_name_to_id: Dict[str, str],
    root: etree._Element,
) -> List[str]:
    # First, we order the pools

    pools_to_tasks = __pools_to_tasks(activity_to_pool_lane)
    ordered_pools = __order_structures(pools_to_tasks, task_name_to_id, root)
    # Then, we order the lanes per pool
    resulting_pool_lane_order = {}
    for pool in ordered_pools:
        lane_data = defaultdict(list)
        for activity, (p, lane) in activity_to_pool_lane.items():
            if p == pool:
                lane_data[lane].append(activity)
        ordered_lanes = __order_structures(lane_data, task_name_to_id, root)
        resulting_pool_lane_order[pool] = ordered_lanes
    return resulting_pool_lane_order


def construct_pools(
    activity_to_pool_lane: Dict[str, Tuple[str, str]],
    model_dimensions: List[int],
    ordering: Dict[str, List[str]],
    padding=50,
    vertical_padding=30,
) -> List[Pool]:
    """
    This function takes the xml output and the pool data and constructs a list of Pool objects accordingly.
    """
    lane_width, lane_height, lane_padding = (
        model_dimensions[0],
        model_dimensions[1],
        model_dimensions[2],
    )
    lane_height += lane_padding
    # get the pools from the xml output
    current_height = 0
    list_of_pools = []
    # get the pool data
    pool_data = {
        activity: pool for activity, (pool, _) in activity_to_pool_lane.items()
    }
    # Sort it according to the ordering
    pool_data = {pool: pool_data for pool in ordering.keys()}
    for pool_name in pool_data.keys():
        # get the name of the pool
        lane_data = {}
        for activity, (p, lane) in activity_to_pool_lane.items():
            if p == pool_name:
                if lane not in lane_data:
                    lane_data[lane] = []
                lane_data[lane].append(activity)
        constructed_lanes = construct_lanes(
            lane_data, lane_width, lane_height, current_height
        )
        # get the up left and down right coordinates of the pool
        up_left = (0, current_height)
        down_right = (
            lane_width + vertical_padding,
            current_height + lane_height * len(lane_data),
        )
        # add the padding accordingly
        list_of_pools.append(Pool(up_left, down_right, pool_name, constructed_lanes))
        current_height += lane_height * len(lane_data) + padding
    return list_of_pools


def construct_lanes(
    lane_data: dict,
    lane_width: int,
    lane_height: int,
    current_height: int = 0,
    vertical_padding: int = 30,
) -> List[Lane]:
    """
    This function takes the xml output and the lane data and returns a list of lanes
    """
    # get the lanes from the xml output
    lanes = []
    # get the lane data
    for lane in lane_data.keys():
        # get the name of the lane
        name = lane
        # get the activities of the lane
        activities = lane_data[lane]
        # get the up left and down right coordinates of the lane
        up_left = (vertical_padding, current_height)
        current_height += lane_height
        down_right = (vertical_padding + lane_width, current_height)
        # create a new lane object
        lanes.append(Lane(up_left, down_right, name, activities))
    return lanes


def __preprocess_start_event(root: etree._Element, element_id):
    # We have to get rid of the     isInterrupting="false" attribute of the start event
    ns = root.nsmap
    BPMN_NS = ns.get("bpmn")
    if not BPMN_NS:
        raise Exception(
            "BPMN namespace not found in the document. Cannot preprocess start event."
        )
    start_event_tag = f"{{{BPMN_NS}}}startEvent"
    start_event = root.find(f".//{start_event_tag}[@id='{element_id}']")
    if start_event is not None and "isInterrupting" in start_event.attrib:
        del start_event.attrib["isInterrupting"]
    return root


def __edit_element_coordinates(
    root: etree._Element, element_id: str, coordinates: Tuple[float, float]
) -> etree._Element:
    """
    This function takes the xml output and the id of the activity and returns the xml output with the new coordinates
    """
    ns = root.nsmap
    if "dc" not in ns and "omgdc" in ns:
        ns["dc"] = ns["omgdc"]

    if "bpmndi" not in ns or ("dc" not in ns and "omgdc" not in ns):
        raise Exception("Required BPMN namespaces not found.")

    query = f'.//bpmndi:BPMNShape[@bpmnElement="{element_id}"]/dc:Bounds'
    bounds_element = root.find(query, namespaces=ns)

    # 3. If the element is found, modify its attributes IN PLACE.
    if bounds_element is not None:
        new_x, new_y = coordinates
        # .set() is the correct way to change an attribute's value.
        # Attribute values must be strings.
        bounds_element.set("x", str(new_x))
        bounds_element.set("y", str(new_y))
    else:
        raise Exception(
            f"Element with id '{element_id}' not found in the XML. No changes made."
        )
    if "Start" in element_id:
        root = __preprocess_start_event(root, element_id)

    return root


def __get_y_coordinates_for_alignment(
    element_id: str, root: etree._Element, lanes
) -> float:
    """
    This function retrieves the y-coordinate for aligning an element based on its neighbors.
    """
    element_coordinates = __get_element_coordinates(root, element_id)
    if not element_coordinates:
        return None
    # returns the middle y-coordinate of the element
    y_coordinate = element_coordinates[1] + element_coordinates[3] / 2
    corresponding_lane = next(
        (lane for lane in lanes if element_id in lane.get_elements()), None
    )
    return y_coordinate, corresponding_lane


def __check_intersection(shape1, shape2):
    """This function takes two shapes and returns True if they intersect"""
    return shape1.intersects(shape2)


def __update_sequence_flow_positions(
    sequence_flow_id, control_flow, root: etree._Element
) -> etree._Element:
    """
    This function takes the control flow and the xml output and returns the xml output with the new coordinates
    """
    ns = root.nsmap

    DI_NS_URI = "http://www.omg.org/spec/DD/20100524/DI"

    dc_ns = ns.get("dc") or ns.get("omgdc")
    xsi_ns = ns.get("xsi")

    if not all([dc_ns, xsi_ns]):
        raise Exception(
            "Error: Required namespaces (dc, xsi) not found in the document."
        )

    # Find the specific BPMNEdge element using an XPath query
    query = f'.//bpmndi:BPMNEdge[@bpmnElement="{sequence_flow_id}"]'
    edge_element = root.find(query, namespaces=ns)

    if edge_element is None:
        raise Exception(f"No match found for sequence flow {sequence_flow_id}")

    saved_attributes = dict(edge_element.attrib)
    edge_element.clear()
    edge_element.attrib.update(saved_attributes)  # Restore attributes after clear

    # Create and append the new waypoint elements.
    for x_val, y_val in control_flow:
        etree.SubElement(
            edge_element,
            f"{{{DI_NS_URI}}}waypoint",
            attrib={f"{{{xsi_ns}}}type": "dc:Point", "x": str(x_val), "y": str(y_val)},
        )
    etree.register_namespace("di", DI_NS_URI)

    return root


def __get_midway_point(
    bounds: Tuple[float, float, float, float]
) -> Tuple[float, float]:
    """
    This function takes the coordinates of the element and returns the middle point
    """
    x1, y1, width, height = bounds
    x = x1 + width / 2
    y = y1 + height / 2
    return (x, y)


def __find_location_of_flow(
    source_bounds: Tuple[float, float, float, float],
    target_bounds: Tuple[float, float, float, float],
) -> str:
    """
    This function takes the coordinates of the source and target and returns the location of the flow
    """
    # Use the centroids
    source_x, source_y = __get_midway_point(source_bounds)
    target_x, target_y = __get_midway_point(target_bounds)
    # Check the direction of the flow
    flow_direction = ""
    if source_x < target_x:
        flow_direction += "r"
    elif source_x > target_x:
        flow_direction += "l"
    if source_y < target_y:
        flow_direction += "u"
    elif source_y > target_y:
        flow_direction += "d"
    return flow_direction


def __get_docking_point(
    bounds: Tuple[float, float, float, float], docking_direction: DockingDirection
) -> Tuple[float, float]:
    """
    This function takes the direction and the coordinates of the element and returns the docking point
    """
    midway_point = __get_midway_point(bounds)
    _, _, width, height = bounds
    if docking_direction == DockingDirection.LEFT:
        return (midway_point[0] - width / 2, midway_point[1])
    elif docking_direction == DockingDirection.RIGHT:
        return (midway_point[0] + width / 2, midway_point[1])
    elif docking_direction == DockingDirection.TOP:
        return (midway_point[0], midway_point[1] + height / 2)
    elif docking_direction == DockingDirection.BOTTOM:
        return (midway_point[0], midway_point[1] - height / 2)


def __check_for_path_intersection(
    path: shapely.MultiLineString,
    shapes: List[shapely.box],
    source_shape: shapely.box,
    target_shape: shapely.box,
) -> bool:
    """
    This function takes the path and the shapes and returns True if the path intersects with any of the shapes
    """
    # Construct a list of all the shapes
    for shape in shapes:
        if __check_intersection(path, shape) or path.touches(shape):
            return True

    # Check if the intersection with the source and target is a point
    intersection_with_src = shapely.intersection(path, source_shape.buffer(-0.05))
    intersection_with_tgt = shapely.intersection(path, target_shape.buffer(-0.05))
    if intersection_with_src.is_empty and intersection_with_tgt.is_empty:
        return False
    elif isinstance(intersection_with_src, shapely.geometry.Point) and isinstance(
        intersection_with_tgt, shapely.geometry.Point
    ):
        return False
    return True


def __construct_s_shaped_flow(
    path_start: Tuple[float, float],
    path_end: Tuple[float, float],
    docking_pt_src: DockingDirection,
    docking_pt_tgt: DockingDirection,
) -> List[List[Tuple[float, float]]]:
    pathways = []
    flow_direction = path_start[0] - path_end[0], path_start[1] - path_end[1]
    left = flow_direction[0] < 0
    # a bit weird but the coordinate system starts at top left (0, 0)
    down = flow_direction[1] < 0
    if (
        docking_pt_src == DockingDirection.LEFT
        and docking_pt_tgt == DockingDirection.RIGHT
        and not left
    ):
        # S-shaped curvature with vertical twist
        aux_pt_1 = (path_start[0] + path_end[0]) / 2, path_start[1]
        aux_pt_2 = (path_start[0] + path_end[0]) / 2, path_end[1]
        pathways.append([aux_pt_1, aux_pt_2])
    elif (
        docking_pt_src == DockingDirection.RIGHT
        and docking_pt_tgt == DockingDirection.LEFT
        and left
    ):
        # S-shaped curvature with vertical twist
        aux_pt_1 = (path_start[0] + path_end[0]) / 2, path_start[1]
        aux_pt_2 = (path_start[0] + path_end[0]) / 2, path_end[1]
        pathways.append([aux_pt_1, aux_pt_2])
    elif (
        docking_pt_src == DockingDirection.TOP
        and docking_pt_tgt == DockingDirection.BOTTOM
        and not down
    ):
        # S-shaped curvature with horizontal twist
        aux_pt_1 = path_start[0], (path_start[1] + path_end[1]) / 2
        aux_pt_2 = path_end[0], (path_start[1] + path_end[1]) / 2
        pathways.append([aux_pt_1, aux_pt_2])
    elif (
        docking_pt_src == DockingDirection.BOTTOM
        and docking_pt_tgt == DockingDirection.TOP
        and down
    ):
        # S-shaped curvature with horizontal twist
        aux_pt_1 = path_start[0], (path_start[1] + path_end[1]) / 2
        aux_pt_2 = path_end[0], (path_start[1] + path_end[1]) / 2
        pathways.append([aux_pt_1, aux_pt_2])
    return pathways


def __construct_c_shaped_flow(
    path_start: Tuple[float, float],
    path_end: Tuple[float, float],
    docking_pt_src: DockingDirection,
    docking_pt_tgt: DockingDirection,
    offset,
) -> List[Tuple[float, float]]:
    pathways = []
    variable_offset = [10, 15, 20]
    if docking_pt_src == docking_pt_tgt:
        if docking_pt_src in {DockingDirection.LEFT, DockingDirection.RIGHT}:
            # C-shaped curvature with vertical
            # We fix the x-coordinate for the vertical case
            x_coord = (
                min(path_start[0], path_end[0])
                if docking_pt_src == DockingDirection.LEFT
                else max(path_start[0], path_end[0])
            )
            for var_offset in variable_offset:
                aux_pt_1 = x_coord + offset + var_offset, path_start[1]
                aux_pt_2 = x_coord + offset + var_offset, path_end[1]
                pathways.append([aux_pt_1, aux_pt_2])
                # Minus outset, too
                aux_pt_1 = x_coord - offset - var_offset, path_start[1]
                aux_pt_2 = x_coord - offset - var_offset, path_end[1]
                pathways.append([aux_pt_1, aux_pt_2])
        else:
            # C-shaped curvature with horizontal twist
            # We fix the y-coordinate for the horizontal case
            y_coord = (
                min(path_start[1], path_end[1])
                if docking_pt_src == DockingDirection.BOTTOM
                else max(path_start[1], path_end[1])
            )
            for var_offset in variable_offset:
                aux_pt_1 = path_start[0], y_coord + offset + var_offset
                aux_pt_2 = path_end[0], y_coord + offset + var_offset
                # Minus outset, too
                pathways.append([aux_pt_1, aux_pt_2])
                aux_pt_1 = path_start[0], y_coord - offset - var_offset
                aux_pt_2 = path_end[0], y_coord - offset - var_offset
                pathways.append([aux_pt_1, aux_pt_2])
    return pathways


def __construct_auxiliary_points(
    path_start: Tuple[float, float],
    path_end: Tuple[float, float],
    docking_pt_src: DockingDirection,
    docking_pt_tgt: DockingDirection,
    offset,
) -> Tuple[List[Tuple[float, float]], List[Tuple[float, float]]]:
    set([docking_pt_src, docking_pt_tgt])
    pathways = []
    if path_start[0] == path_end[0] or path_start[1] == path_end[1]:
        if (
            docking_pt_src == DockingDirection.LEFT
            and docking_pt_tgt == DockingDirection.RIGHT
        ) or (
            docking_pt_src == DockingDirection.BOTTOM
            and docking_pt_tgt == DockingDirection.TOP
        ):
            # We allow only LEFT-RIGHT and BOTTOM-TOP straight connections
            # Otherwise, we might end up with some weird cases, e.g., BOTTOM-BOTTOM straight lines
            pathways.append([None, None])

    # Case 1 L-Shaped: Vertical Exit -> Horizontal Entry
    if docking_pt_src in {
        DockingDirection.BOTTOM,
        DockingDirection.TOP,
    } and docking_pt_tgt in {DockingDirection.LEFT, DockingDirection.RIGHT}:

        aux_pt = (path_start[0], path_end[1])
        pathways.append([aux_pt, None])

    # Case 2 L-Shaped: Horizontal Exit -> Vertical Entry
    elif docking_pt_src in {
        DockingDirection.LEFT,
        DockingDirection.RIGHT,
    } and docking_pt_tgt in {DockingDirection.BOTTOM, DockingDirection.TOP}:

        aux_pt = (path_end[0], path_start[1])
        pathways.append([None, aux_pt])

    pathways.extend(
        __construct_c_shaped_flow(
            path_start, path_end, docking_pt_src, docking_pt_tgt, offset
        )
    )

    pathways.extend(
        __construct_s_shaped_flow(path_start, path_end, docking_pt_src, docking_pt_tgt)
    )

    return pathways


def __get_docking_point_name(
    docking_point: Tuple[float, float], figure: Tuple[float, float, float, float]
) -> str:
    """
    This function takes the docking point and the figure and returns the name of the docking point
    """
    # Get the coordinates of the figure
    x1, y1, width, height = figure
    if docking_point[0] == x1 + width / 2 and docking_point[1] == y1:
        return DockingDirection.TOP
    elif docking_point[0] == x1 + width / 2 and docking_point[1] == y1 + height:
        return DockingDirection.BOTTOM
    elif docking_point[0] == x1 and docking_point[1] == y1 + height / 2:
        return DockingDirection.LEFT
    elif docking_point[0] == x1 + width and docking_point[1] == y1 + height / 2:
        return DockingDirection.RIGHT
    else:
        raise ValueError("Provided point is not a docking point")


def __turn_points_into_multi_linestring(
    points: List[Tuple[float, float]]
) -> shapely.MultiLineString:
    """
    This function takes a list of points (tuples) and returns a MultiLineString object
    """
    # First, turn them into LineStrings
    lines = []
    for i in range(len(points) - 1):
        line = shapely.LineString([points[i], points[i + 1]])
        lines.append(line)
    # Now, turn them into a MultiLineString
    return shapely.MultiLineString(lines)


def __check_for_flow_connectivity(
    current_path: List, prev_paths: List[List[Tuple[float, float]]]
) -> bool:
    incoming_docking_points = [path[-1] for path in prev_paths]
    outgoing_docking_points = [path[0] for path in prev_paths]
    return (
        current_path[0] in incoming_docking_points
        or current_path[-1] in outgoing_docking_points
    )


def __prioritize_paths(paths: List, source, target) -> List:
    dx = __get_midway_point(target)[0] - __get_midway_point(source)[0]
    dy = __get_midway_point(target)[1] - __get_midway_point(source)[1]

    best_start = None
    if abs(dy) > abs(dx):
        # Vertical movement is dominant
        best_start = DockingDirection.BOTTOM if dy > 0 else DockingDirection.TOP
    elif abs(dx) > abs(dy):
        best_start = DockingDirection.RIGHT if dx > 0 else DockingDirection.LEFT
    else:
        # Prioritize vertical movement
        best_start = DockingDirection.BOTTOM if dy > 0 else DockingDirection.TOP

    def get_score(path):

        # Visual complexity - number of turns
        not path[1]
        total_path_intersection = path[2]
        num_turns = len(path[0])

        # Check misalignment
        actual_start = __get_docking_point_name(path[0][0], source)
        is_misaligned = actual_start != best_start

        # Total path Length (Manhattan distance)
        length = 0
        for i in range(len(path[0]) - 1):
            length += abs(path[0][i][0] - path[0][i + 1][0]) + abs(
                path[0][i][1] - path[0][i + 1][1]
            )

        return (
            num_turns,  # Should be simple
            is_misaligned,
            length,
            total_path_intersection,  # Should have minimal intersection with other paths
        )

    return sorted(paths, key=get_score)


def __check_for_intersection_with_other_paths(
    current_path: shapely.MultiLineString,
    prev_paths: List[List[Tuple[float, float]]],
):
    # Should return a float with total intersection length
    total_intersection_length = 0.0
    for path in prev_paths:
        other_path = __turn_points_into_multi_linestring(path)
        intersection = current_path.intersection(other_path)
        if not intersection.is_empty:
            total_intersection_length += intersection.length
    return total_intersection_length


def __construct_possible_paths(
    source_docking_points: List[Tuple[float, float]],
    target_docking_points: List[Tuple[float, float]],
    shapes_to_consider: List,
    source_coords: Tuple[float, float, float, float],
    target_coords: Tuple[float, float, float, float],
    prev_paths: List[List[Tuple[float, float]]],
) -> List[shapely.LineString]:
    """
    This function takes the docking points of the source and target and returns the first suitable path
    That does not intersect with any of the shapes
    """
    paths = product(source_docking_points, target_docking_points)
    possible_full_paths = []
    backup_paths = []
    source_shape = shapely.box(
        source_coords[0],
        source_coords[1],
        source_coords[0] + source_coords[2],
        source_coords[1] + source_coords[3],
    )
    target_shape = shapely.box(
        target_coords[0],
        target_coords[1],
        target_coords[0] + target_coords[2],
        target_coords[1] + target_coords[3],
    )
    remaining_shapes = [
        shape
        for shape in shapes_to_consider
        if not shapely.equals(shape, source_shape)
        and not shapely.equals(shape, target_shape)
    ]
    # Get target height for outset
    randomness = random.randrange(3, 9) / 10
    offset = randomness * max(target_coords[3], source_coords[3])
    for src, target in paths:
        # Check if the path intersects with any of the shapes
        src_coords_connection = __get_docking_point(source_coords, src)
        tgt_coords_connection = __get_docking_point(target_coords, target)
        pathways = __construct_auxiliary_points(
            src_coords_connection, tgt_coords_connection, src, target, offset
        )
        for i in range(len(pathways)):
            points_to_add = []
            for point in pathways[i]:
                if point is not None:
                    points_to_add.append(point)
            constructed_path = [
                src_coords_connection,
                *points_to_add,
                tgt_coords_connection,
            ]
            len(constructed_path)
            multilane_path = __turn_points_into_multi_linestring(constructed_path)
            backup_paths.append(
                (
                    constructed_path,
                    __check_for_intersection_with_other_paths(
                        multilane_path, prev_paths
                    ),
                )
            )
            if not __check_for_path_intersection(
                multilane_path, remaining_shapes, source_shape, target_shape
            ):
                # If it is exactly of length 2, it means it is a straight line
                # We can take it directly
                """
                if path_length == 2:
                    print("Found straight line path, returning it directly")
                    return constructed_path
                """
                # If it does not intersect, we can add it to the list

                possible_full_paths.append(
                    (
                        constructed_path,
                        __check_for_flow_connectivity(constructed_path, prev_paths),
                        __check_for_intersection_with_other_paths(
                            multilane_path, prev_paths
                        ),
                    )
                )

    if len(possible_full_paths) != 0:
        # First, sort them by the boolean argument, then by length of the path
        possible_full_paths = __prioritize_paths(
            possible_full_paths, source_coords, target_coords
        )
        return possible_full_paths[0][0]
    print("No suitable path found, returning backup path")
    # We will choose the path with the least intersection with other paths
    backup_paths = sorted(backup_paths, key=lambda x: x[1])
    return backup_paths[0][0]


def __construct_possible_directions(
    flow: str,
) -> Tuple[List[DockingDirection], List[DockingDirection]]:
    if flow == "r" or flow == "l":

        possible_directions_1 = [
            DockingDirection.RIGHT,
            DockingDirection.TOP,
            DockingDirection.BOTTOM,
        ]
        possible_directions_2 = [
            DockingDirection.LEFT,
            DockingDirection.TOP,
            DockingDirection.BOTTOM,
        ]
        if flow == "r":
            possible_directions_src = possible_directions_1
            possible_directions_tgt = possible_directions_2
        else:
            possible_directions_src = possible_directions_2
            possible_directions_tgt = possible_directions_1
    elif flow == "u" or flow == "d":
        possible_directions_1 = [
            DockingDirection.TOP,
            DockingDirection.LEFT,
            DockingDirection.RIGHT,
        ]
        possible_directions_2 = [
            DockingDirection.BOTTOM,
            DockingDirection.LEFT,
            DockingDirection.RIGHT,
        ]
        if flow == "u":
            possible_directions_src = possible_directions_1
            possible_directions_tgt = possible_directions_2
        else:
            possible_directions_src = possible_directions_2
            possible_directions_tgt = possible_directions_1
    elif flow == "ru" or flow == "ur":
        possible_directions_src = [
            DockingDirection.RIGHT,
            DockingDirection.TOP,
        ]
        possible_directions_tgt = [
            DockingDirection.LEFT,
            DockingDirection.BOTTOM,
        ]
    elif flow == "rd" or flow == "dr":
        possible_directions_src = [
            DockingDirection.RIGHT,
            DockingDirection.BOTTOM,
        ]
        possible_directions_tgt = [
            DockingDirection.LEFT,
            DockingDirection.TOP,
        ]
    elif flow == "lu" or flow == "ul":
        possible_directions_src = [
            DockingDirection.LEFT,
            DockingDirection.TOP,
        ]
        possible_directions_tgt = [
            DockingDirection.RIGHT,
            DockingDirection.BOTTOM,
        ]
    elif flow == "ld" or flow == "dl":
        possible_directions_src = [
            DockingDirection.LEFT,
            DockingDirection.BOTTOM,
        ]
        possible_directions_tgt = [
            DockingDirection.RIGHT,
            DockingDirection.TOP,
        ]

    return possible_directions_src, possible_directions_tgt


def connect_points(
    source_coords,
    target_coords,
    shapes: List[shapely.box],
    prev_paths: List[List[Tuple[float, float]]],
) -> Tuple[Tuple[float, float]]:
    """
    This function takes the source and target shapes and returns the docking point
    """
    # flow = __find_location_of_flow(source_coords, target_coords)
    # possible_docking_points_src, possible_docking_points_tgt = __construct_possible_directions(flow)
    possible_docking_points_src = [
        DockingDirection.LEFT,
        DockingDirection.RIGHT,
        DockingDirection.TOP,
        DockingDirection.BOTTOM,
    ]
    possible_docking_points_tgt = [
        DockingDirection.LEFT,
        DockingDirection.RIGHT,
        DockingDirection.TOP,
        DockingDirection.BOTTOM,
    ]
    path = __construct_possible_paths(
        possible_docking_points_src,
        possible_docking_points_tgt,
        shapes,
        source_coords,
        target_coords,
        prev_paths,
    )

    return path


def __identify_sequence_flows(root: etree._Element) -> List[str]:
    """
    This function takes the xml output and returns the ids of the sequence flows
    """
    flows = []
    ns = root.nsmap
    if "bpmn" not in ns:
        raise Exception(
            "BPMN namespace not found in the document. Cannot process sequence flows."
        )

    flow_elements = root.xpath("//bpmn:sequenceFlow", namespaces=ns)

    for flow in flow_elements:
        flow_id = flow.get("id")
        source_ref = flow.get("sourceRef")
        target_ref = flow.get("targetRef")
        flows.append((flow_id, source_ref, target_ref))
    connections_dict = {
        seq_flow: (source, target) for seq_flow, source, target in flows
    }
    return connections_dict


def __handle_sequence_flows(root: etree._Element, shapes: List[object]):
    """
    This function takes the xml output and the lanes and returns the xml output with the sequence flows
    """
    # Get the sequence flows
    connections_dict = __identify_sequence_flows(root)
    # Get rid of previous sequence flows so we can add the new ones w/o any issues
    prev_paths = []
    message_flows = []
    # sort the connection dict based on the following heuristic
    # First, try to connect gateways with gateways, then tasks with tasks, then events with events, then the rest
    connections_dict = dict(
        sorted(
            connections_dict.items(),
            key=lambda item: (
                0
                if "Gateway" in item[1][0] and "Gateway" in item[1][1]
                else 1
                if "Task" in item[1][0] and "Task" in item[1][1]
                else 2
                if ("Event" in item[1][0] and "Event" in item[1][1])
                else 3
            ),
        )
    )
    for seq_flow, (source, target) in connections_dict.items():
        src_coords = __get_element_coordinates(root, source)
        tgt_coords = __get_element_coordinates(root, target)
        if "Intermediate" in source and "Intermediate" in target:
            # Intermediate events are not connected with sequence flows, but with message flows
            message_flows.append((seq_flow, source, target))
        # Should be embedded here though
        path = connect_points(src_coords, tgt_coords, shapes, prev_paths)
        prev_paths.append(path)
        root = __update_sequence_flow_positions(seq_flow, path, root)
    return root, message_flows


def __align_tasks(
    lanes: List[Lane], root: str, task_dict: Dict[str, str]
) -> Tuple[str, List[str]]:
    aligned_elements = []
    for lane in lanes:
        # get the up left and down right coordinates of the lane
        up_left = lane.get_up_left()
        activities = lane.get_activities()
        for activity in activities:
            # get the id of the activity
            id_activity = task_dict[activity]
            aligned_elements.append(id_activity)
            lane.add_element(id_activity)
            # edit the coordinates of the task
            task_coordinates = __get_element_coordinates(root, id_activity)
            new_coordinates = (task_coordinates[0], task_coordinates[1] + up_left[1])
            root = __edit_element_coordinates(root, id_activity, new_coordinates)

    return root, aligned_elements


def __check_gateway_type(element_id: str, flows: Dict[str, Tuple[str, str]]) -> str:
    """
    This function checks if the gateway is diverging or converging
    """
    incoming, outgoing = 0, 0
    for _, (source, target) in flows.items():
        if source == element_id:
            outgoing += 1
        elif target == element_id:
            incoming += 1
    if incoming > outgoing:
        return "converging"
    elif outgoing > incoming:
        return "diverging"
    else:
        return "mixed"


def __identify_nearest_aligned_element(
    element_id: str,
    aligned_pool_elements: List[str],
    seq_flows: Dict[str, Tuple[str, str]],
    root: etree._Element,
) -> str:
    """
    This function identifies the nearest aligned element in the pool.
    """
    predecessors, successors = [], []
    for _, (source, target) in seq_flows.items():
        if source == element_id and target in aligned_pool_elements:
            successors.append(target)
        elif target == element_id and source in aligned_pool_elements:
            predecessors.append(source)

    # We will apply a heuristic here
    if "Gateway" in element_id:
        # We have to check if it is diverging or converging
        match_type = __check_gateway_type(element_id, seq_flows)

        match match_type:
            case "converging":
                if len(predecessors) > 0:
                    return predecessors[0]
            case "diverging":
                if len(successors) > 0:
                    return successors[0]
            case _:
                pass

    neighbors = predecessors + successors
    if len(neighbors) > 0:
        return neighbors[0]
    else:
        min_x_distance = float("inf")
        nearest_element = None
        current_element_coordinates = __get_element_coordinates(root, element_id)
        for aligned_element in aligned_pool_elements:
            element_coordinates = __get_element_coordinates(root, aligned_element)
            if element_coordinates is None:
                continue
            x_distance = abs(current_element_coordinates[0] - element_coordinates[0])
            if x_distance < min_x_distance:
                min_x_distance = x_distance
                nearest_element = aligned_element
        # Identify to which lane it belongs

        return nearest_element


def __align_gateways(
    root: etree._Element,
    coloring: Dict[str, str],
    aligned_elements: List[str],
    lanes: List[Lane],
    seq_flows: Dict[str, Tuple[str, str]],
) -> Tuple[str, List[str]]:
    # This function aligns the gateways in the BPMN diagram.
    prev_iteration = aligned_elements.copy()
    element_lane_mapping = {}
    i = 0
    while set(prev_iteration) != set(aligned_elements) or i < 3:
        prev_iteration = aligned_elements.copy()
        for el_id, color in coloring.items():
            if "Gateway" not in el_id:
                continue
            # Otherwise, we need to locate the nearest element within the same pool

            aligned_elements_within_pool = [
                el
                for el in aligned_elements
                if coloring.get(el) == color and el != el_id
            ]
            if len(aligned_elements_within_pool) == 0:
                continue
            nearest_pool_element = __identify_nearest_aligned_element(
                el_id, aligned_elements_within_pool, seq_flows, root
            )
            if nearest_pool_element is not None:
                # Get the coordinates of the nearest element
                nearest_coordinates = __get_element_coordinates(
                    root, nearest_pool_element
                )
                # Get the lane of the nearest element
                nearest_y, corresponding_lane = __get_y_coordinates_for_alignment(
                    nearest_pool_element, root, lanes
                )
                current_element_coordinates = __get_element_coordinates(root, el_id)
                try:
                    y_coordinate = (
                        current_element_coordinates[1]
                        + corresponding_lane.get_up_left()[1]
                    )
                    if (
                        y_coordinate > corresponding_lane.get_down_right()[1] - 20
                        or y_coordinate < corresponding_lane.get_up_left()[1] + 20
                    ):
                        continue
                except AttributeError:
                    continue

                root = __edit_element_coordinates(
                    root, el_id, (current_element_coordinates[0], y_coordinate)
                )
                element_lane_mapping[el_id] = corresponding_lane
                if el_id not in aligned_elements:
                    aligned_elements.append(el_id)
        i += 1
    for el_id, lane in element_lane_mapping.items():
        lane.add_element(el_id)
    return root


def __align_events(
    root: etree._Element,
    coloring: Dict[str, str],
    aligned_elements: List[str],
    lanes: List[Lane],
    seq_flows: Dict[str, Tuple[str, str]],
) -> Tuple[str, List[str]]:
    for el_id, color in coloring.items():
        if el_id in aligned_elements:
            continue
        # Otherwise, we need to locate the nearest element within the same pool

        aligned_elements_within_pool = [
            el for el in aligned_elements if coloring.get(el) == color and el != el_id
        ]
        nearest_pool_element = __identify_nearest_aligned_element(
            el_id, aligned_elements_within_pool, seq_flows, root
        )
        if nearest_pool_element is not None:
            # Get the coordinates of the nearest element
            __get_element_coordinates(root, nearest_pool_element)
            # Get the lane of the nearest element
            nearest_y, corresponding_lane = __get_y_coordinates_for_alignment(
                nearest_pool_element, root, lanes
            )
            current_element_coordinates = __get_element_coordinates(root, el_id)
            y_coordinate = (
                current_element_coordinates[1] + corresponding_lane.get_up_left()[1]
            )
            root = __edit_element_coordinates(
                root, el_id, (current_element_coordinates[0], y_coordinate)
            )
            corresponding_lane.add_element(el_id)
            if el_id not in aligned_elements:
                aligned_elements.append(el_id)
    return root


def __align_elements(
    root: etree._Element,
    coloring: Dict[str, str],
    aligned_elements: List[str],
    lanes: List[Lane],
) -> etree._Element:
    # This is a simple heuristic to ensure that tasks are aligned first, then gateways, then events
    # This ensures that the start and end events are connected to their respective predecessors/successors
    # in a more natural way.
    ordered_elements = sorted(
        coloring.keys(),
        key=lambda x: (0 if "Task" in x else (1 if "Gateway" in x else 2)),
    )
    seq_flows = __identify_sequence_flows(root)
    coloring = {el: coloring[el] for el in ordered_elements}
    root = __align_gateways(root, coloring, aligned_elements, lanes, seq_flows)
    root = __align_events(root, coloring, aligned_elements, lanes, seq_flows)
    return root


def __create_shapes(elements, root: etree._Element):
    """This function takes the xml output and returns a list of shapely objects"""
    list_of_shapes = []
    for element in elements:
        # get the coordinates of the element
        element_coordinates = __get_element_coordinates(root, element)
        # create a shapely object
        shape = shapely.box(
            element_coordinates[0],
            element_coordinates[1],
            element_coordinates[0] + element_coordinates[2],
            element_coordinates[1] + element_coordinates[3],
        )

        # add the shape to the plot
        list_of_shapes.append(shape)
    return list_of_shapes


def __change_intermediate_throw_events_to_message_events(root: etree.Element) -> None:
    """
    This function changes all intermediate throw events to message events in the BPMN XML.
    """
    BPMN_NS = root.nsmap.get("bpmn")
    if not BPMN_NS:
        raise ValueError("BPMN namespace not found in the XML root.")

    for event in root.findall(f".//{{{BPMN_NS}}}intermediateThrowEvent"):
        # Should have       <bpmn:messageEventDefinition id="MessageEventDefinition_0soq2t8" /> within it
        # add it at the bottom of the event
        etree.SubElement(
            event,
            f"{{{BPMN_NS}}}messageEventDefinition",
            attrib={"id": f'MessageEventDefinition_{hash(event.get("id"))}'},
        )
    return root


def __add_collaboration(
    root: etree.Element, pools: List[Pool], msg_flows: List[Tuple[str, str, str]]
) -> etree.Element:
    """
    This function restructures a flat BPMN into a collaboration with pools and lanes,
    correctly handling message flows and updating the diagram plane.
    """
    BPMN_NS = root.nsmap.get("bpmn")
    BPMNDI_NS = root.nsmap.get("bpmndi")

    plane = root.find(f".//{{{BPMNDI_NS}}}BPMNPlane")

    if not BPMN_NS or not BPMNDI_NS:
        raise ValueError("BPMN or BPMNDI namespace not found in the XML root.")

    source_process = root.find(f".//{{{BPMN_NS}}}process")
    if source_process is None:
        raise ValueError("No <bpmn:process> element found in the source XML.")

    collaboration = etree.Element(
        f"{{{BPMN_NS}}}collaboration", attrib={"id": "Collaboration_1"}
    )

    # Create a set of message flow IDs for efficient lookup. These are flows
    # that should NOT be treated as sequence flows.
    msg_flow_ids = {flow_data[0] for flow_data in msg_flows}

    element_to_new_process_map = {}
    new_processes = []

    for pool in pools:
        process_id = f"Process_{hash(pool.get_name())}"
        etree.SubElement(
            collaboration,
            f"{{{BPMN_NS}}}participant",
            attrib={
                "id": f"Participant_{process_id}",
                "name": pool.get_name(),
                "processRef": process_id,
            },
        )
        new_process = etree.Element(
            f"{{{BPMN_NS}}}process",
            attrib={"id": process_id, "name": pool.get_name(), "isExecutable": "false"},
        )
        create_visual_shape(
            plane,
            f"Participant_{process_id}",
            [pool.get_up_left(), pool.get_down_right()],
            root,
        )

        laneset = etree.SubElement(new_process, f"{{{BPMN_NS}}}laneSet")

        for lane in pool.get_lanes():
            lane_element = etree.SubElement(
                laneset,
                f"{{{BPMN_NS}}}lane",
                attrib={
                    "id": f"Lane_{hash(lane.get_name())}_{hash(pool.get_name())}",
                    "name": lane.get_name(),
                },
            )
            create_visual_shape(
                plane,
                f"Lane_{hash(lane.get_name())}_{hash(pool.get_name())}",
                [lane.get_up_left(), lane.get_down_right()],
                root,
            )
            for element_id in lane.get_elements():
                flow_node_ref = etree.SubElement(
                    lane_element, f"{{{BPMN_NS}}}flowNodeRef"
                )
                flow_node_ref.text = element_id
                element_to_new_process_map[element_id] = new_process
        new_processes.append(new_process)

    for element in list(source_process):
        element_id = element.get("id")

        # Check if it's a flow element (Task, Event, Gateway)
        if element_id in element_to_new_process_map:
            destination_process = element_to_new_process_map[element_id]
            destination_process.append(element)

        # Check if it's a flow connector
        elif element.tag == f"{{{BPMN_NS}}}sequenceFlow":
            # If the ID is in our set of message flows, DISCARD IT.
            # It will be recreated as a <messageFlow> later.
            if element_id in msg_flow_ids:
                continue  # Skip and do nothing with this element

            # Otherwise, it's a true sequence flow. Move it to the correct process.
            else:
                source_ref = element.get("sourceRef")
                if source_ref in element_to_new_process_map:
                    destination_process = element_to_new_process_map[source_ref]
                    destination_process.append(element)

    for proc in new_processes:
        root.append(proc)

    for flow_id, source_id, target_id in msg_flows:
        etree.SubElement(
            collaboration,
            f"{{{BPMN_NS}}}messageFlow",
            attrib={"id": flow_id, "sourceRef": source_id, "targetRef": target_id},
        )

    root.append(collaboration)
    root.remove(source_process)

    root = __change_intermediate_throw_events_to_message_events(root)
    # Find the BPMNPlane and make sure it references the collaboration
    plane = root.find(f".//{{{BPMNDI_NS}}}BPMNPlane")
    if plane is not None:
        plane.set("bpmnElement", collaboration.get("id"))

    return root


def build_lanes_only(root: etree._Element, lanes: List[Lane]) -> str:
    """
    This function takes the xml output and the lanes and returns the xml output with the lanes.
    """


def build_pools_with_collaboration(
    root: etree._Element, pools: List[Pool], msg_flows: List[Tuple[str, str, str]]
) -> str:
    """
    This function takes the xml output and the pools and returns the xml output with the pools.
    """

    __add_collaboration(root, pools, msg_flows)

    # Return the modified XML as a string
    return _to_string(root)


def create_visual_shape(
    plane: etree.Element,
    logical_id: str,
    coordinates: Tuple[Tuple[int, int], Tuple[int, int]],
    root,
):
    """
    Creates a BPMNShape with its Bounds on the BPMNPlane.

    Args:
        plane: The lxml element for the <bpmndi:BPMNPlane>.
        logical_id: The ID of the <bpmn:participant> or <bpmn:lane> to link to.
        coordinates: A tuple containing ((x1, y1), (x2, y2)) for top-left and bottom-right corners.
        ns: A dictionary of required namespaces ('bpmndi', 'omgdc').
    """
    BPMNDI_NS = root.nsmap.get("bpmndi")
    OMGDC_NS = root.nsmap.get("omgdc")

    top_left, bottom_right = coordinates
    x1, y1 = top_left
    x2, y2 = bottom_right

    bounds_attributes = {
        "x": str(x1),
        "y": str(y1),
        "width": str(x2 - x1),
        "height": str(y2 - y1),
    }

    # Create the <bpmndi:BPMNShape> for the pool or lane
    shape = etree.SubElement(
        plane,
        f"{{{BPMNDI_NS}}}BPMNShape",
        attrib={
            "id": f"{logical_id}_di",  # Standard convention: logical_id + _di
            "bpmnElement": logical_id,
            "isHorizontal": "true",  # Pools and horizontal lanes are marked this way
        },
    )

    # Create the <omgdc:Bounds> element inside the shape
    etree.SubElement(shape, f"{{{OMGDC_NS}}}Bounds", attrib=bounds_attributes)


def postprocess_diagram(root: etree._Element, pools: List[Pool]) -> etree._Element:
    current_y_cursor = None
    for pool in pools:
        current_y_cursor = (
            pool.get_up_left()[1]
            if current_y_cursor is None
            else min(current_y_cursor, pool.get_up_left()[1])
        )
        if current_y_cursor != pool.get_up_left()[1]:
            # Update pool position
            pool.set_up_left((pool.get_up_left()[0], current_y_cursor))

        lanes = pool.get_lanes()
        lanes.sort(key=lambda l: l.get_up_left()[1])

        for lane in lanes:
            elements_in_lane = lane.get_elements()

            # 1. Find the bounding box of the elements currently in this lane
            current_elements_coords = {}
            for element_id in elements_in_lane:
                coords = __get_element_coordinates(root, element_id)
                if coords:
                    # x, y, width, height
                    current_elements_coords[element_id] = coords

            if not current_elements_coords:
                raise ValueError(
                    f"No elements found in lane '{lane.get_name()}' for alignment."
                )

            # Calculate content height
            min_ele_y = min(c[1] for c in current_elements_coords.values())
            max_ele_y = max(c[1] + c[3] for c in current_elements_coords.values())
            content_height = max_ele_y - min_ele_y

            # 2. Calculate how much we need to shift these elements to sit at current_y_cursor + PADDING
            # We want the top-most element to be at (current_y_cursor + PADDING)
            target_top_y = current_y_cursor + PADDING_LANES
            shift_amount = target_top_y - min_ele_y

            # 3. Apply shift to all elements in this lane
            for element_id, (x, y, w, h) in current_elements_coords.items():
                new_y = y + shift_amount
                root = __edit_element_coordinates(root, element_id, (x, new_y))

            # 4. Resize the Lane
            new_lane_height = content_height + (PADDING_LANES * 2)
            lane.set_up_left((lane.get_up_left()[0], current_y_cursor))
            lane.set_down_right(
                (lane.get_down_right()[0], current_y_cursor + new_lane_height)
            )

            # 5. Move cursor down for the next lane
            current_y_cursor += new_lane_height

        # 6. Finally, resize the Pool to fit the stacked lanes
        pool.set_down_right((pool.get_down_right()[0], current_y_cursor))
        current_y_cursor += PADDING_POOLS  # Add pool padding after all lanes

    return root



from pm4py.objects.process_tree.obj import Operator as PTOperator, ProcessTree

from powl.objects.obj import (
    OperatorPOWL,
    POWL,
    Sequence,
    SilentTransition,
    StrictPartialOrder,
    Transition,
)


def apply(tree: ProcessTree) -> POWL:

    nodes = []

    for c in tree.children:
        nodes.append(apply(c))

    if tree.operator is None:
        if tree.label is not None:
            powl = Transition(label=tree.label)
        else:
            powl = SilentTransition()
    elif tree.operator == PTOperator.XOR:
        powl = OperatorPOWL(PTOperator.XOR, nodes)
    elif tree.operator == PTOperator.LOOP:
        powl = OperatorPOWL(PTOperator.LOOP, nodes)
    elif tree.operator == PTOperator.PARALLEL:
        powl = StrictPartialOrder(nodes=nodes)
    elif tree.operator == PTOperator.SEQUENCE:
        powl = Sequence(nodes)
    else:
        raise Exception("Unsupported process tree!")

    powl = powl.simplify()

    return powl



from pm4py.algo.discovery.inductive.fall_through.activity_once_per_trace import (
    ActivityOncePerTraceUVCL,
)

from powl.discovery.total_order_based.inductive.fall_through.activity_concurrent import (
    POWLActivityConcurrentUVCL,
)


class POWLActivityOncePerTraceUVCL(
    ActivityOncePerTraceUVCL, POWLActivityConcurrentUVCL
):
    pass



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List as TList, Optional, Union

from pm4py.objects.process_tree.obj import Operator, ProcessTree

from powl.objects.BinaryRelation import BinaryRelation


class POWL(ProcessTree, ABC):
    @abstractmethod
    def simplify_using_frequent_transitions(self) -> "POWL":
        return self

    @abstractmethod
    def simplify(self) -> "POWL":
        return self

    def __str__(self):
        return self.__repr__()

    @abstractmethod
    def reduce_silent_transitions(self, add_empty_paths=True) -> "POWL":
        return self


class Transition(POWL):
    transition_id: int = 0

    def __init__(
        self, label: Optional[str] = None, organization=None, role=None
    ) -> None:
        super().__init__()
        self._label = label
        self._organization = organization
        self._role = role
        self._identifier = Transition.transition_id
        Transition.transition_id = Transition.transition_id + 1

    def __repr__(self) -> str:
        return f"Transition(label={self._label}, id={self._identifier})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Transition):
            return self._label == other._label and self._identifier == other._identifier
        return False

    def __copy__(self):
        return Transition(self._label)

    def __deepcopy__(self, memo):
        return Transition(self._label)

    def equal_content(self, other: object) -> bool:
        if isinstance(other, Transition):
            return self._label == other._label
        return False

    def set_organization(self, organization):
        self._organization = organization

    def set_role(self, role):
        self._role = role

    def __hash__(self) -> int:
        return self._identifier

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Transition):
            if self.label and other.label and self.label < other.label:
                return self.label < other.label
            return self._identifier < other._identifier
        elif isinstance(other, OperatorPOWL):
            return True
        elif isinstance(other, StrictPartialOrder):
            return True
        return NotImplemented

    def simplify_using_frequent_transitions(self) -> "Transition":
        return self

    def simplify(self) -> "Transition":
        return self

    def reduce_silent_transitions(self, add_empty_paths=True) -> "Transition":
        return self


class SilentTransition(Transition):
    def __init__(self) -> None:
        super().__init__(label=None)

    def __copy__(self):
        return SilentTransition()

    def __deepcopy__(self, memo):
        return SilentTransition()


class FrequentTransition(Transition):
    def __init__(
        self, label, min_freq: Union[str, int], max_freq: Union[str, int]
    ) -> None:
        self.skippable = False
        self.selfloop = False
        if min_freq == 0:
            self.skippable = True
        if max_freq == "-":
            self.selfloop = True
        self.activity = label
        if self.skippable or self.selfloop:
            label = str(label) + "\n" + "[" + str(min_freq) + "," + str(max_freq) + "]"

        super().__init__(label=label)

    def simplify(self):
        raise Exception(
            "Not allowed! You cannot call the simplify function on powl models annotated with frequency tags."
        )

    def __repr__(self):
        return f"FrequentTransition(activity={self.activity}, skippable={self.skippable}, selfloop={self.selfloop})"

    def set_skippable(self, skip):
        self.skippable = skip
        self.__update_label()

    def set_selfloop(self, loop):
        self.selfloop = loop
        self.__update_label()

    def __update_label(self):
        if self.skippable:
            min_freq = 0
        else:
            min_freq = 1
        if self.selfloop:
            max_freq = "-"
        else:
            max_freq = 1
        self.label = (
            str(self.activity) + "\n" + "[" + str(min_freq) + "," + str(max_freq) + "]"
        )

    def __copy__(self):
        if self.skippable:
            min_freq = 0
        else:
            min_freq = 1
        if self.selfloop:
            max_freq = "-"
        else:
            max_freq = 1
        ft = FrequentTransition(self.activity, min_freq, max_freq)
        ft._organization = self._organization
        ft._role = self._role
        return ft

    def __deepcopy__(self, memo):
        if self.skippable:
            min_freq = 0
        else:
            min_freq = 1
        if self.selfloop:
            max_freq = "-"
        else:
            max_freq = 1
        ft = FrequentTransition(self.activity, min_freq, max_freq)
        ft._organization = self._organization
        ft._role = self._role
        return ft


class StrictPartialOrder(POWL):
    def __init__(self, nodes: TList[POWL]) -> None:
        super().__init__()
        self.operator = None
        self._set_order(nodes)
        self.additional_information = None

    def __repr__(self):
        return f"StrictPartialOrder({self.order.nodes})"

    def _set_order(self, nodes: TList[POWL]) -> None:
        self.order = BinaryRelation(nodes)

    def get_order(self) -> BinaryRelation:
        return self.order

    def _set_children(self, children: TList[POWL]) -> None:
        self.order.nodes = children

    def get_children(self) -> TList[POWL]:
        return self.order.nodes

    def __lt__(self, other: object) -> bool:
        if isinstance(other, StrictPartialOrder):
            return self.__repr__() < other.__repr__()
        elif isinstance(other, OperatorPOWL):
            return False
        elif isinstance(other, Transition):
            return False
        return NotImplemented

    partial_order = property(get_order, _set_order)
    children = property(get_children, _set_children)

    def equal_content(self, other: object) -> bool:
        if not isinstance(other, StrictPartialOrder):
            return False

        ordered_nodes_1 = sorted(list(self.order.nodes))
        ordered_nodes_2 = sorted(list(other.order.nodes))
        if len(ordered_nodes_1) != len(ordered_nodes_2):
            return False
        for i in range(len(ordered_nodes_1)):
            source_1 = ordered_nodes_1[i]
            source_2 = ordered_nodes_2[i]
            if not source_1.equal_content(source_2):
                return False
            for j in range(len(ordered_nodes_1)):
                target_1 = ordered_nodes_1[j]
                target_2 = ordered_nodes_2[j]
                if self.order.is_edge(source_1, target_1) and not other.order.is_edge(
                    source_2, target_2
                ):
                    return False
                if not self.order.is_edge(source_1, target_1) and other.order.is_edge(
                    source_2, target_2
                ):
                    return False
        return True

    def simplify_using_frequent_transitions(self) -> "StrictPartialOrder":
        new_nodes = {
            node: node.simplify_using_frequent_transitions() for node in self.children
        }
        res = StrictPartialOrder(list(new_nodes.values()))
        for node_1 in self.children:
            for node_2 in self.children:
                if self.partial_order.is_edge(node_1, node_2):
                    res.partial_order.add_edge(new_nodes[node_1], new_nodes[node_2])

        return res

    def reduce_silent_transitions(self, add_empty_paths=True) -> "StrictPartialOrder":
        new_nodes_map = {
            node: node.reduce_silent_transitions(add_empty_paths)
            for node in self.children
            if not isinstance(node, SilentTransition)
        }
        return self.map_nodes(new_nodes_map)

    def simplify(self) -> "StrictPartialOrder":
        simplified_nodes = {}
        sub_nodes = {}
        start_nodes = {}
        end_nodes = {}

        def connected(node):
            for node2 in self.children:
                if self.partial_order.is_edge(
                    node, node2
                ) or self.partial_order.is_edge(node2, node):
                    return True
            return False

        for node_1 in self.children:
            simplified_node = node_1.simplify()
            if isinstance(simplified_node, StrictPartialOrder):

                if not connected(node_1):
                    sub_nodes[node_1] = simplified_node
                else:
                    s_nodes = simplified_node.order.get_start_nodes()
                    e_nodes = simplified_node.order.get_end_nodes()
                    if len(s_nodes) == 1 and len(e_nodes) == 1:
                        sub_nodes[node_1] = simplified_node
                        start_nodes[node_1] = list(s_nodes)[0]
                        end_nodes[node_1] = list(e_nodes)[0]
                    else:
                        simplified_nodes[node_1] = simplified_node
            else:
                simplified_nodes[node_1] = simplified_node

        new_nodes = list(simplified_nodes.values())
        for po, simplified_po in sub_nodes.items():
            new_nodes = new_nodes + list(simplified_po.children)
        res = StrictPartialOrder(new_nodes)
        for node_1 in self.children:
            for node_2 in self.children:
                if self.partial_order.is_edge(node_1, node_2):
                    if (
                        node_1 in simplified_nodes.keys()
                        and node_2 in simplified_nodes.keys()
                    ):
                        res.partial_order.add_edge(
                            simplified_nodes[node_1], simplified_nodes[node_2]
                        )
                    elif node_1 in simplified_nodes.keys():
                        res.partial_order.add_edge(
                            simplified_nodes[node_1], start_nodes[node_2]
                        )
                    elif node_2 in simplified_nodes.keys():
                        res.partial_order.add_edge(
                            end_nodes[node_1], simplified_nodes[node_2]
                        )
                    else:
                        res.partial_order.add_edge(
                            end_nodes[node_1], start_nodes[node_2]
                        )
        for po, simplified_po in sub_nodes.items():
            for node_1 in simplified_po.children:
                for node_2 in simplified_po.children:
                    if simplified_po.partial_order.is_edge(node_1, node_2):
                        res.partial_order.add_edge(node_1, node_2)
        return res

    def add_edge(self, source, target):
        return self.order.add_edge(source, target)

    def map_nodes(self, mapping):
        res = StrictPartialOrder(list(mapping.values()))
        for node_1, new_node_1 in mapping.items():
            for node_2, new_node_2 in mapping.items():
                if self.partial_order.is_edge(node_1, node_2):
                    res.partial_order.add_edge(new_node_1, new_node_2)
        return res


class Sequence(StrictPartialOrder):
    def __init__(self, nodes: TList[POWL]) -> None:
        super().__init__(nodes)
        for i in range(len(nodes)):
            for j in range(i + 1, len(nodes)):
                self.partial_order.add_edge(nodes[i], nodes[j])

    def simplify_using_frequent_transitions(self) -> "StrictPartialOrder":
        new_children = []
        for node in self.children:
            s_node = node.simplify_using_frequent_transitions()
            if isinstance(s_node, Sequence):
                new_children = new_children + s_node.children
            else:
                new_children = new_children + [s_node]
        res = Sequence(new_children)
        return res

    def reduce_silent_transitions(self, add_empty_paths=True) -> "Sequence":
        new_nodes = [
            node.reduce_silent_transitions(add_empty_paths)
            for node in self.children
            if not isinstance(node, SilentTransition)
        ]
        return Sequence(new_nodes)


class OperatorPOWL(POWL):
    operator_id: int = 0
    def __init__(self, operator: Operator, children: TList[POWL]) -> None:
        super().__init__()
        self.operator = operator
        self.children = children
        self._identifier = OperatorPOWL.operator_id
        OperatorPOWL.operator_id = OperatorPOWL.operator_id + 1

    def __repr__(self):
        return f"{self.operator}({self.children})"

    def __lt__(self, other: object) -> bool:
        if isinstance(other, OperatorPOWL):
            return self.__repr__() < other.__repr__()
        elif isinstance(other, Transition):
            return False
        elif isinstance(other, StrictPartialOrder):
            return True
        return NotImplemented

    def equal_content(self, other: object) -> bool:
        if not isinstance(other, OperatorPOWL):
            return False

        if self.operator != other.operator:
            return False

        ordered_nodes_1 = sorted(list(self.children))
        ordered_nodes_2 = sorted(list(other.children))
        if len(ordered_nodes_1) != len(ordered_nodes_2):
            return False
        for i in range(len(ordered_nodes_1)):
            node_1 = ordered_nodes_1[i]
            node_2 = ordered_nodes_2[i]
            if not node_1.equal_content(node_2):
                return False
        return True

    def reduce_silent_transitions(self, add_empty_paths=True) -> POWL:

        new_children = [
            node.reduce_silent_transitions(add_empty_paths) for node in self.children
        ]

        if self.operator == Operator.XOR:
            new_children_no_silent = [
                c for c in new_children if not isinstance(c, SilentTransition)
            ]
            if len(new_children_no_silent) < len(new_children) - 1:
                new_children = new_children_no_silent + [SilentTransition()]
                if len(new_children) == 1:
                    return new_children[0]

        return OperatorPOWL(operator=self.operator, children=new_children)

    def simplify_using_frequent_transitions(self) -> POWL:
        if self.operator is Operator.XOR and len(self.children) == 2:
            child_0 = self.children[0].simplify_using_frequent_transitions()
            child_1 = self.children[1].simplify_using_frequent_transitions()

            if isinstance(child_0, SilentTransition) and isinstance(
                child_1, SilentTransition
            ):
                return child_0

            if isinstance(child_0, SilentTransition):
                if isinstance(child_1, FrequentTransition):
                    child_1.set_skippable(True)
                    return child_1
                elif isinstance(child_1, Transition):
                    return FrequentTransition(
                        label=child_1.label, min_freq=0, max_freq=1
                    )
            if isinstance(child_1, SilentTransition):
                if isinstance(child_0, FrequentTransition):
                    child_0.set_skippable(True)
                    return child_0
                elif isinstance(child_0, Transition):
                    return FrequentTransition(
                        label=child_0.label, min_freq=0, max_freq=1
                    )

        if self.operator is Operator.LOOP and len(self.children) == 2:
            child_0 = self.children[0]
            child_1 = self.children[1]
            if isinstance(child_0, Transition) and isinstance(
                child_1, SilentTransition
            ):
                return FrequentTransition(label=child_0.label, min_freq=1, max_freq="-")
            elif isinstance(child_1, Transition) and isinstance(
                child_0, SilentTransition
            ):
                return FrequentTransition(label=child_1.label, min_freq=0, max_freq="-")

        return OperatorPOWL(
            self.operator,
            [child.simplify_using_frequent_transitions() for child in self.children],
        )

    def simplify(self) -> "OperatorPOWL":
        if self.operator is Operator.XOR and len(self.children) == 2:
            child_0 = self.children[0]
            child_1 = self.children[1]

            def merge_with_children(child0, child1):
                if (
                    isinstance(child0, SilentTransition)
                    and isinstance(child1, OperatorPOWL)
                    and child1.operator is Operator.LOOP
                ):
                    if isinstance(child1.children[0], SilentTransition):
                        return OperatorPOWL(
                            Operator.LOOP, [n.simplify() for n in child1.children]
                        )
                    elif isinstance(child1.children[1], SilentTransition):
                        return OperatorPOWL(
                            Operator.LOOP,
                            list(reversed([n.simplify() for n in child1.children])),
                        )

                return None

            res = merge_with_children(child_0, child_1)
            if res is not None:
                return res

            res = merge_with_children(child_1, child_0)
            if res is not None:
                return res

        return OperatorPOWL(
            self.operator, [child.simplify() for child in self.children]
        )


class DecisionGraph(POWL):
    """
    A DecisionGraph is a POWL model defined over a set of nodes (each node is a POWL model)
    together with a binary relation (order) over these nodes, augmented with two artificial
    nodes: a start node and an end node.

    In the decision graph, each node represents a group (or branch) of activities (or submodels)
    and the binary relation encodes the allowed ordering between these nodes.
    """

    def __init__(
        self, order: BinaryRelation, start_nodes, end_nodes, empty_path=False
    ) -> None:
        super().__init__()
        self.operator = None
        self.children = [n for n in order.nodes]
        self.start_nodes = start_nodes
        self.end_nodes = end_nodes
        # if not all(isinstance(node, POWL) for node in nodes):
        #     raise Exception("The nodes of the decision graph must be POWL models!")
        if not start_nodes or not set(start_nodes).issubset(order.nodes):
            raise Exception(
                "Start nodes must be a non-empty subset of the nodes of the relation!"
            )
        if not end_nodes or not set(end_nodes).issubset(order.nodes):
            raise Exception(
                "End nodes must be a non-empty subset of the nodes of the relation!"
            )
        self.start = StartNode()
        self.end = EndNode()
        order.add_node(self.start)
        order.add_node(self.end)
        for node in start_nodes:
            order.add_edge(self.start, node)
        for node in end_nodes:
            order.add_edge(node, self.end)
        if empty_path:
            order.add_edge(self.start, self.end)

        self.order = order

    def __repr__(self):
        return f"DecisionGraph({self.children})"

    def simplify(self) -> POWL:
        if len(self.children) == 1:
            child_0 = self.children[0]
            skippable = self.order.is_edge(self.start, self.end)
            repeatable = self.order.is_edge(child_0, child_0)

            if skippable:
                if repeatable:
                    return OperatorPOWL(
                        Operator.LOOP, [SilentTransition(), child_0]
                    ).simplify()
                else:
                    if isinstance(child_0, DecisionGraph):
                        child_0.empty_path = True
                        child_0.order.add_edge(child_0.start, child_0.end)
                        return child_0.simplify()
                    else:
                        return OperatorPOWL(
                            Operator.XOR, [SilentTransition(), child_0]
                        ).simplify()

            elif repeatable:
                return OperatorPOWL(
                    Operator.LOOP, [child_0, SilentTransition()]
                ).simplify()

            else:
                return child_0.simplify()

        else:
            new_dg = self

            seq = new_dg.__group_start_seq()
            if seq:
                return seq.simplify()

            seq = new_dg.__group_end_seq()
            if seq:
                return seq.simplify()

            res = new_dg.__group_pure_seq()
            if len(res.children) < len(new_dg.children):
                return res.simplify()

            new_children_map = {}
            for child in new_dg.children:
                s_child = child.simplify()
                new_children_map[child] = s_child
            return new_dg.__apply_mapping(new_children_map)

    def simplify_using_frequent_transitions(self) -> POWL:
        if len(self.children) == 1:
            child_0 = self.children[0]

            if isinstance(child_0, Transition):
                skippable = self.order.is_edge(self.start, self.end)
                repeatable = self.order.is_edge(child_0, child_0)

                min_freq = 0 if skippable else 1
                max_freq = "-" if repeatable else 1

                if skippable or repeatable:
                    return FrequentTransition(
                        label=child_0.label, min_freq=min_freq, max_freq=max_freq
                    )
                else:
                    return child_0

        new_children_map = {}
        edges_to_remove = set()
        for child in self.children:
            s_child = child.simplify_using_frequent_transitions()
            if isinstance(s_child, Transition):
                preset = self.order.get_preset(child)
                postset = self.order.get_postset(child)
                if all(
                    self.order.is_edge(pre, post) for pre in preset for post in postset
                ):
                    for pre in preset:
                        for post in postset:
                            edges_to_remove.add((pre, post))
                    if child in self.start_nodes:
                        self.start_nodes = [
                            x for x in self.start_nodes if x not in postset
                        ]
                    if child in self.end_nodes:
                        self.end_nodes = [x for x in self.end_nodes if x not in preset]
                    if isinstance(s_child, FrequentTransition):
                        s_child.set_skippable(True)
                    else:
                        s_child = FrequentTransition(
                            label=child.label, min_freq=0, max_freq=1
                        )
            new_children_map[child] = s_child
        new_dg = self.__apply_mapping(new_children_map, edges_to_remove)

        old_size = len(new_dg.children) + 1
        while len(new_dg.children) < old_size:
            old_size = len(new_dg.children)
            new_dg = new_dg.__group_pure_seq()

        seq1 = new_dg.__group_start_seq()
        start_nodes = []
        if seq1:
            start_nodes = seq1.children[:-1]
            new_dg = seq1.children[-1]

        seq2 = new_dg.__group_end_seq()
        end_nodes = []
        if seq2:
            end_nodes = seq2.children[1:]
            new_dg = seq2.children[0]

        new_node = new_dg
        if len(new_dg.children) == 1:
            child_0 = new_dg.children[0]
            if new_dg.order.is_edge(new_dg.start, new_dg.end):
                if (
                    isinstance(child_0, OperatorPOWL)
                    and child_0.operator is Operator.LOOP
                ):
                    if isinstance(child_0.children[0], SilentTransition):
                        new_node = OperatorPOWL(Operator.LOOP, child_0.children)
                    elif isinstance(child_0.children[1], SilentTransition):
                        new_node = OperatorPOWL(
                            Operator.LOOP, list(reversed(child_0.children))
                        )
                elif isinstance(child_0, DecisionGraph):
                    child_0.empty_path = True
                    child_0.order.add_edge(child_0.start, child_0.end)
                    new_node = child_0
            else:
                new_node = child_0

        if len(start_nodes) > 0 or len(end_nodes) > 0:
            if isinstance(new_node, Sequence):
                return Sequence(start_nodes + new_node.children + end_nodes)
            else:
                return Sequence(start_nodes + [new_node] + end_nodes)
        else:
            return new_node

    def __apply_mapping(self, mapping, edges_to_remove=None) -> "DecisionGraph":
        if edges_to_remove is None:
            edges_to_remove = set()
        res = BinaryRelation(list(set(mapping.values())))
        for src in self.children:
            for tgt in self.children:
                if self.order.is_edge(src, tgt) and (src, tgt) not in edges_to_remove:
                    new_src = mapping[src]
                    new_tgt = mapping[tgt]
                    if new_src != new_tgt:
                        res.add_edge(new_src, new_tgt)
        new_start_nodes = list({mapping[child] for child in self.start_nodes})
        new_end_nodes = list({mapping[child] for child in self.end_nodes})
        empty_path = (
            self.order.is_edge(self.start, self.end)
            and not (self.start, self.end) in edges_to_remove
        )
        return DecisionGraph(res, new_start_nodes, new_end_nodes, empty_path)

    def __create_mapping(self, old_children, new_child):
        mapping = {}
        for key in self.children:
            if key in old_children:
                mapping[key] = new_child
            else:
                mapping[key] = key
        return mapping

    # def __group_pure_xor(self):
    #     new_dg = deepcopy(self)
    #     visited = set()
    #     for child in list(self.children):
    #         if child not in visited:
    #             visited.add(child)
    #             post1 = new_dg.order.get_postset(child)
    #             pre1 = new_dg.order.get_preset(child)
    #             if len(pre1) > 1 or len(post1) > 1:
    #                 candidates_for_xor = [child2 for child2 in new_dg.children if
    #                                       pre1 == new_dg.order.get_preset(child2) and post1 == new_dg.order.get_postset(
    #                                           child2)]
    #                 if len(candidates_for_xor) > 1:
    #                     visited.update(candidates_for_xor)
    #
    #                     new_rel = BinaryRelation(candidates_for_xor)
    #                     empty_path = False
    #                     edges_to_remove = None
    #                     if all(new_dg.order.is_edge(pre, post) for pre in pre1 for post in post1):
    #                         empty_path = True
    #                         edges_to_remove = [(pre, post) for pre in pre1 for post in post1]
    #                     dg_child = DecisionGraph(new_rel, candidates_for_xor, candidates_for_xor, empty_path)
    #                     mapping = new_dg.__create_mapping(candidates_for_xor, dg_child)
    #                     new_dg = new_dg.__apply_mapping(mapping, edges_to_remove)
    #     return new_dg

    def __group_pure_seq(self):
        for child in list(self.children):
            for child2 in self.children:
                post1 = self.order.get_postset(child)
                pre2 = self.order.get_preset(child2)
                if pre2 == {child} and post1 == {child2}:
                    seq = Sequence([child, child2])
                    mapping = self.__create_mapping({child, child2}, seq)
                    new_dg = self.__apply_mapping(mapping)
                    return new_dg.__group_pure_seq()
        return self

    def __group_start_seq(self):
        start_list = []
        current_dg = self
        while (
            len(current_dg.children) > 1
            and len(current_dg.start_nodes) == 1
            and not current_dg.order.is_edge(current_dg.start, current_dg.end)
            and current_dg.order.get_preset(current_dg.start_nodes[0])
            == {current_dg.start}
        ):
            start = current_dg.start_nodes[0]
            start_list.append(start)
            postset = current_dg.order.get_postset(start)
            new_start_nodes = list(postset - {current_dg.end})
            new_children = [n for n in current_dg.children if n != start]
            new_end_nodes = [n for n in current_dg.end_nodes if n != start]
            new_order = BinaryRelation(new_children)
            for c1 in new_children:
                for c2 in new_children:
                    if current_dg.order.is_edge(c1, c2):
                        new_order.add_edge(c1, c2)
            empty_path = current_dg.end in postset
            current_dg = DecisionGraph(
                new_order, new_start_nodes, new_end_nodes, empty_path
            )
        if len(start_list) > 0:
            seq = Sequence(start_list + [current_dg])
            return seq
        return None

    def __group_end_seq(self):
        end_list = []
        current_dg = self
        while (
            len(current_dg.children) > 1
            and len(current_dg.end_nodes) == 1
            and not current_dg.order.is_edge(current_dg.start, current_dg.end)
            and current_dg.order.get_postset(current_dg.end_nodes[0])
            == {current_dg.end}
        ):
            end = current_dg.end_nodes[0]
            end_list = [end] + end_list
            pretset = current_dg.order.get_preset(end)
            new_end_nodes = list(pretset - {current_dg.start})
            new_children = [n for n in current_dg.children if n != end]
            new_start_nodes = [n for n in current_dg.start_nodes if n != end]
            new_order = BinaryRelation(new_children)
            for c1 in new_children:
                for c2 in new_children:
                    if current_dg.order.is_edge(c1, c2):
                        new_order.add_edge(c1, c2)
            empty_path = current_dg.start in pretset
            current_dg = DecisionGraph(
                new_order, new_start_nodes, new_end_nodes, empty_path
            )
        if len(end_list) > 0:
            seq = Sequence([current_dg] + end_list)
            return seq
        return None

    def reduce_silent_transitions(self, add_empty_paths=True) -> "POWL":
        graph_copy = deepcopy(self)

        mapping = {
            node: node.reduce_silent_transitions(add_empty_paths)
            for node in graph_copy.children
        }
        order_nodes = [n for n in graph_copy.order.nodes]

        for node, new_node in mapping.items():
            if isinstance(new_node, SilentTransition):
                order_nodes.remove(node)
                for n1 in order_nodes:
                    for n2 in order_nodes:
                        if graph_copy.order.is_edge(
                            n1, node
                        ) and graph_copy.order.is_edge(node, n2):
                            graph_copy.order.add_edge(n1, n2)

        mapping = {key: value for key, value in mapping.items() if key in order_nodes}

        if len(mapping.keys()) == 0:
            return SilentTransition()
        else:
            new_start = [
                value
                for key, value in mapping.items()
                if graph_copy.order.is_edge(graph_copy.start, key)
            ]
            new_end = [
                value
                for key, value in mapping.items()
                if graph_copy.order.is_edge(key, graph_copy.end)
            ]
            skip = graph_copy.order.is_edge(graph_copy.start, graph_copy.end)
            new_order = BinaryRelation(list(mapping.values()))
            for node_1 in mapping.keys():
                for node_2 in mapping.keys():
                    if graph_copy.order.is_edge(node_1, node_2):
                        new_order.add_edge(mapping[node_1], mapping[node_2])
            if skip and not add_empty_paths:
                old_skip = self.order.is_edge(self.start, self.end)
                if not old_skip:
                    graph_copy.order.remove_edge(graph_copy.start, graph_copy.end)
                    skip = False
            return DecisionGraph(new_order, new_start, new_end, skip)

    def map_nodes(self, mapping):
        assert all(child in mapping.keys() for child in self.children)
        new_children = list(mapping.values())
        new_order = BinaryRelation(new_children)
        for node_1, new_node_1 in mapping.items():
            for node_2, new_node_2 in mapping.items():
                if self.order.is_edge(node_1, node_2):
                    new_order.add_edge(new_node_1, new_node_2)
        new_start = [mapping[n] for n in self.start_nodes]
        new_end = [mapping[n] for n in self.end_nodes]
        empty_path = self.order.is_edge(self.start, self.end)
        return DecisionGraph(new_order, new_start, new_end, empty_path)


class StartNode:
    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return "⊕start"


class EndNode:
    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return "⊕end"



import os
from abc import ABC
from enum import Enum
from itertools import combinations
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.empty_traces import EmptyTracesUVCL
from pm4py.algo.discovery.inductive.variants.imf import IMFParameters
from pm4py.objects.process_tree.obj import Operator
from pm4py.util import constants, exec_utils

from powl.discovery.total_order_based.inductive.base_case.factory import BaseCaseFactory
from powl.discovery.total_order_based.inductive.cuts.factory import CutFactory
from powl.discovery.total_order_based.inductive.fall_through.empty_traces import (
    POWLEmptyTracesUVCL,
)
from powl.discovery.total_order_based.inductive.fall_through.factory import (
    FallThroughFactory,
)
from powl.discovery.total_order_based.inductive.utils.filtering import (
    filter_most_frequent_variants,
    filter_most_frequent_variants_with_decreasing_factor,
    FILTERING_THRESHOLD,
    FILTERING_TYPE,
    FilteringType,
)
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.general_utils.dfg_frequency_filtering import filter_dfg_noise_keep_activities_and_repair
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import (
    DecisionGraph,
    OperatorPOWL,
    POWL,
    Sequence,
    StrictPartialOrder,
)

T = TypeVar("T", bound=IMDataStructureUVCL)


class Parameters(Enum):
    MULTIPROCESSING = "multiprocessing"


class IMBasePOWL(ABC, Generic[T]):
    def __init__(self, parameters: Optional[Dict[str, Any]] = None):
        if parameters is None:
            parameters = {}

        enable_multiprocessing = exec_utils.get_param_value(
            Parameters.MULTIPROCESSING,
            parameters,
            constants.ENABLE_MULTIPROCESSING_DEFAULT,
        )

        if enable_multiprocessing:
            from multiprocessing import Manager, Pool

            self._pool = Pool(os.cpu_count() - 1)
            self._manager = Manager()
            self._manager.support_list = []
        else:
            self._pool = None
            self._manager = None

    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.TREE

    def empty_traces_cut(self) -> Type[EmptyTracesUVCL]:
        return POWLEmptyTracesUVCL

    def enable_dfg_fall_through(self) -> bool:
        return False

    def apply(
        self,
        obj: T,
        parameters: Optional[Dict[str, Any]] = None,
        second_iteration: bool = False,
    ) -> POWL:

        noise_threshold = exec_utils.get_param_value(
            IMFParameters.NOISE_THRESHOLD, parameters, 0.0
        )

        empty_traces = self.empty_traces_cut().apply(obj, parameters)
        if empty_traces is not None:
            number_original_traces = sum(y for y in obj.data_structure.values())
            number_filtered_traces = sum(
                y for y in empty_traces[1][-1].data_structure.values()
            )

            if (
                number_original_traces - number_filtered_traces
                > noise_threshold * number_original_traces
            ):
                return self._recurse(empty_traces[0], empty_traces[1], parameters)
            else:
                obj = empty_traces[1][-1]

        powl = self.apply_base_cases(obj, parameters)
        if powl is not None:
            return powl

        cut = self.find_cut(obj, parameters)
        if cut is not None:
            powl = self._recurse(cut[0], cut[1], parameters=parameters)

        if powl is not None:
            return powl

        if FILTERING_TYPE in parameters.keys():
            filtering_type = parameters[FILTERING_TYPE]
            if filtering_type not in FilteringType:
                raise KeyError("Invalid FILTERING_TYPE: " + str(filtering_type))

            if filtering_type is FilteringType.DFG_FREQUENCY:
                if not second_iteration:
                    noise_threshold = exec_utils.get_param_value(
                        IMFParameters.NOISE_THRESHOLD, parameters, 0.0
                    )
                    filtered_ds = filter_dfg_noise_keep_activities_and_repair(obj, noise_threshold)
                    tree = self.apply(
                        filtered_ds, parameters=parameters, second_iteration=True
                    )
                    if tree is not None:
                        return tree

            elif filtering_type is FilteringType.DYNAMIC:
                filtered_log = filter_most_frequent_variants(obj.data_structure)
                if len(filtered_log.data_structure) > 0:
                    return self.apply(filtered_log, parameters=parameters)

            elif filtering_type is FilteringType.DECREASING_FACTOR:
                if FILTERING_THRESHOLD in parameters.keys():
                    t = parameters[FILTERING_THRESHOLD]
                    if isinstance(t, float) and 0 <= t < 1:
                        t = [t]
                    if isinstance(t, list):
                        for factor in t:
                            if factor > 0:
                                filtered_log = filter_most_frequent_variants_with_decreasing_factor(
                                    obj.data_structure, decreasing_factor=factor
                                )
                                if len(filtered_log.data_structure) == 0:
                                    break
                                elif len(filtered_log.data_structure) < len(
                                    obj.data_structure
                                ):
                                    return self.apply(
                                        filtered_log, parameters=parameters
                                    )
                    else:
                        raise KeyError("Invalid filtering threshold!")
            else:
                raise KeyError("Invalid filtering type!")

        ft = self.fall_through(obj, parameters)
        return self._recurse(ft[0], ft[1], parameters=parameters)

    def apply_base_cases(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[POWL]:
        return BaseCaseFactory.apply_base_cases(obj, parameters=parameters)

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        return CutFactory.find_cut(obj, parameters=parameters)

    def fall_through(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Tuple[POWL, List[T]]:
        return FallThroughFactory.fall_through(
            obj,
            self._pool,
            self._manager,
            enable_dfg_fall_through=self.enable_dfg_fall_through(),
            parameters=parameters,
        )

    def _recurse(
        self, powl: POWL, objs: List[T], parameters: Optional[Dict[str, Any]] = None
    ):
        children = [self.apply(obj, parameters=parameters) for obj in objs]
        if isinstance(powl, StrictPartialOrder):
            if isinstance(powl, Sequence):
                return Sequence(children)
            powl_new = StrictPartialOrder(children)
            for i, j in combinations(range(len(powl.children)), 2):
                if powl.order.is_edge_id(i, j):
                    powl_new.order.add_edge(children[i], children[j])
                elif powl.order.is_edge_id(j, i):
                    powl_new.order.add_edge(children[j], children[i])
            return powl_new
        elif isinstance(powl, DecisionGraph):
            new_order = BinaryRelation(children)
            for i, j in combinations(range(len(powl.children)), 2):
                if powl.order.is_edge(objs[i], objs[j]):
                    new_order.add_edge(children[i], children[j])
                if powl.order.is_edge(objs[j], objs[i]):
                    new_order.add_edge(children[j], children[i])
            start_nodes = [
                children[i]
                for i in range(len(powl.children))
                if objs[i] in powl.start_nodes
            ]
            end_nodes = [
                children[i]
                for i in range(len(powl.children))
                if objs[i] in powl.end_nodes
            ]
            empty_path = powl.order.is_edge(powl.start, powl.end)
            return DecisionGraph(
                new_order, start_nodes, end_nodes, empty_path=empty_path
            )
        elif isinstance(powl, OperatorPOWL):
            if powl.operator == Operator.LOOP and len(children) > 2:
                new_child = OperatorPOWL(Operator.XOR, children[1:])
                children = [children[0], new_child]
            return OperatorPOWL(powl.operator, children)

        else:
            raise Exception("Unsupported POWL type!")
